# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/22 18:16
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: image_document.py
from .document import Document

try:
    import pandas as pd
except ImportError:
    raise ImportError(
        "The functionality you are trying to use requires the pandas and pandas package. "
        "You can install it by running: pip install pandas"
    )


class GraphDocument(Document):
    """The basic class for an ImageDocument.

    Attributes:
        graph_data: A pandas dataframe contents all results of neo4j cypher
    """
    graph_data: pd.DataFrame = None



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/9 18:06
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: log_sink_manager.py
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase
from agentuniverse.base.util.logging.log_sink.log_sink import LogSink


@singleton
class LogSinkManager(ComponentManagerBase[LogSink]):
    """A singleton manager class of the DocProcessor."""

    def __init__(self):
        super().__init__(ComponentEnum.LOG_SINK)
        


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/24 21:19
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: rag_template.py
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.template.rag_agent_template import RagAgentTemplate


class SummarizeRagAgentTemplate(RagAgentTemplate):

    def input_keys(self) -> list[str]:
        return ['input', 'summarize_content']

    def output_keys(self) -> list[str]:
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        agent_input['input'] = input_object.get_data('input')
        agent_input['summarize_content'] = input_object.get_data('summarize_content')
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        return {**agent_result, 'output': agent_result['output']}



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/22 19:15
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: view_file_tool.py

import os
import json

from agentuniverse.agent.action.tool.tool import Tool, ToolInput


class ViewFileTool(Tool):
    def execute(self, tool_input: ToolInput) -> str:
        file_path = tool_input.get_data("file_path")
        start_line = tool_input.get_data("start_line", 0)
        end_line = tool_input.get_data("end_line", None)

        if not file_path or not os.path.isfile(file_path):
            return json.dumps({
                "error": f"File not found: {file_path}",
                "status": "error"
            })

        try:
            with open(file_path, 'r', encoding='utf-8') as file:
                all_lines = file.readlines()
            if end_line is None:
                end_line = len(all_lines)
            start_line = max(0, start_line)
            end_line = min(len(all_lines), end_line)

            content_lines = all_lines[start_line:end_line]
            content = ''.join(content_lines)
            return json.dumps({
                "file_path": file_path,
                "content": content,
                "start_line": start_line,
                "end_line": end_line - 1 if end_line > 0 else 0,
                "total_lines": len(all_lines),
                "status": "success"
            })

        except Exception as e:
            return json.dumps({
                "error": str(e),
                "file_path": file_path,
                "status": "error"
            })



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/22 17:00
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: run_command_tool.py

import os
import json
import time
import threading
import subprocess
from enum import Enum
from typing import Dict, Optional
from dataclasses import dataclass

from agentuniverse.agent.action.tool.tool import Tool, ToolInput


class CommandStatus(Enum):
    RUNNING = "running"
    COMPLETED = "completed"
    ERROR = "error"


@dataclass
class CommandResult:
    thread_id: int
    stdout: str
    stderr: str
    start_time: float
    status: CommandStatus
    exit_code: Optional[int] = None
    end_time: Optional[float] = None

    @property
    def duration(self) -> float:
        if self.end_time is None:
            return time.time() - self.start_time
        return self.end_time - self.start_time

    @property
    def message(self) -> str:
        # Truncate stdout and stderr if they are too long
        max_output_length = 2000
        truncated_stdout = self._truncate_output(
            self.stdout, max_output_length)
        truncated_stderr = self._truncate_output(
            self.stderr, max_output_length)

        result_dict = {
            'thread_id': self.thread_id,
            'status': self.status.value,
            'stdout': truncated_stdout,
            'stderr': truncated_stderr,
            'exit_code': self.exit_code,
            'duration': self.duration
        }
        return json.dumps(result_dict)

    def _truncate_output(self, output: str, max_length: int) -> str:
        """Truncate output to keep beginning and end, removing middle when too long"""
        if not output or len(output) <= max_length:
            return output
        half_length = max_length // 2
        return output[:half_length] + "\n... [truncated output] ...\n" + output[-half_length:]


_command_results: Dict[int, CommandResult] = {}


class RunCommandTool(Tool):
    """
    Tool for executing shell commands either synchronously or asynchronously.
    """

    def execute(self, tool_input: ToolInput) -> str:
        command = tool_input.get_data("command")
        cwd = tool_input.get_data("cwd", os.getcwd())
        blocking = tool_input.get_data("blocking", True)
        return self._run_command(command, cwd, blocking)

    def _run_command(self, command: str, cwd: str, blocking: bool = True) -> str:
        result = CommandResult(
            thread_id=threading.get_ident(),
            status=CommandStatus.RUNNING,
            stdout="",
            stderr="",
            start_time=time.time(),
        )

        def __run() -> None:
            try:
                process = subprocess.Popen(
                    command,
                    cwd=cwd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                    shell=True
                )

                stdout, stderr = process.communicate()
                exit_code = process.returncode

                result.stdout = stdout
                result.stderr = stderr
                result.exit_code = exit_code
                result.end_time = time.time()
                result.status = CommandStatus.COMPLETED if exit_code == 0 else CommandStatus.ERROR

            except Exception as e:
                result.stderr = str(e)
                result.end_time = time.time()
                result.status = CommandStatus.ERROR

            _command_results[result.thread_id] = result

        if blocking:
            __run()
        else:
            thread = threading.Thread(target=__run)
            thread.start()
            result.thread_id = thread.ident
            _command_results[result.thread_id] = result

        return result.message


def get_command_result(thread_id: int) -> Optional[CommandResult]:
    return _command_results.get(thread_id)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
import json
# @Time    : 2025/2/1 09:50
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: contextual_iteration_agent_template.py
from typing import Optional
from langchain_core.output_parsers import StrOutputParser

from agentuniverse.agent.agent import Agent
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.base.util.agent_util import assemble_memory_input, assemble_memory_output
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.llm.llm import LLM
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.base.util.logging.logging_util import LOGGER


class ContextualIterationAgentTemplate(AgentTemplate):
    iteration: int = 0
    continue_prompt_version: Optional[str] = None
    if_loop_prompt_version: Optional[str] = None

    def input_keys(self) -> list[str]:
        return ['input']

    def output_keys(self) -> list[str]:
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        agent_input['input'] = input_object.get_data('input')
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        return {**agent_result, 'output': agent_result['output']}

    def customized_execute(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM, prompt: Prompt,
                           **kwargs) -> dict:
        assemble_memory_input(memory, agent_input)
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile,
                          agent_input)
        conversation_history = []
        combined_res = ''

        # first llm call step
        lc_prompt_template = prompt.as_langchain()
        prompt_input_dict = {key: agent_input[key] for key in lc_prompt_template.input_variables
                             if key in agent_input}

        chain = lc_prompt_template | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        res = self.invoke_chain(chain, agent_input, input_object, **kwargs)

        if self.iteration < 1:
            return {**agent_input, 'output': res}

        # process first result
        combined_res = res
        conversation_history.append({
            'user': lc_prompt_template.format(**prompt_input_dict),
            'assistant': res
        })
        agent_input['chat_history'] = json.dumps(conversation_history, ensure_ascii=False)

        # build interation chain
        continue_prompt: Prompt = super(AgentTemplate, self).process_prompt(
            agent_input, prompt_version=self.continue_prompt_version)
        continue_chain = continue_prompt.as_langchain() | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        if self.iteration > 1:
            if_loop_prompt: Prompt = super(AgentTemplate,
                                            self).process_prompt(
                agent_input, prompt_version=self.if_loop_prompt_version)
            if_loop_prompt_chain = if_loop_prompt.as_langchain() | llm.as_langchain_runnable(
                self.agent_model.llm_params()) | StrOutputParser()

        for i in range(self.iteration):
            continue_res = self.invoke_chain(continue_chain, agent_input, input_object, **kwargs)
            res = res + '\n' + continue_res

            if i==self.iteration-1:
                break
            # judge whether loop
            if not self.if_loop_prompt_version:
                continue
            loop_res = self.invoke_chain(if_loop_prompt_chain, agent_input, input_object, **kwargs)
            if not self.if_loop(loop_res):
                break

            lc_prompt_template = continue_prompt.as_langchain()
            prompt_input_dict = {key: agent_input[key] for key in
                                 lc_prompt_template.input_variables
                                 if key in agent_input}
            conversation_history.append({
                'user': lc_prompt_template.format(**prompt_input_dict),
                'assistant': continue_res
            })
            agent_input['chat_history'] = json.dumps(conversation_history,
                                                     ensure_ascii=False)

        assemble_memory_output(memory=memory,
                               agent_input=agent_input,
                               content=f"Human: {agent_input.get('input')}, AI: {res}")
        return {**agent_input, 'output': res}

    async def customized_async_execute(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM,
                                       prompt: Prompt, **kwargs) -> dict:

        assemble_memory_input(memory, agent_input)
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile,
                          agent_input)
        conversation_history = []
        combined_res = ''

        # first llm call step
        lc_prompt_template = prompt.as_langchain()
        prompt_input_dict = {key: agent_input[key] for key in
                             lc_prompt_template.input_variables
                             if key in agent_input}

        chain = lc_prompt_template | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        res = await self.async_invoke_chain(chain, agent_input, input_object, **kwargs)

        if self.iteration < 1:
            return {**agent_input, 'output': res}

        # process first result
        combined_res = res
        conversation_history.append({
            'user': lc_prompt_template.format(**prompt_input_dict),
            'assistant': res
        })
        agent_input['chat_history'] = json.dumps(conversation_history,
                                                 ensure_ascii=False)

        # build interation chain
        continue_prompt: Prompt = super(AgentTemplate, self).process_prompt(
            agent_input, prompt_version=self.continue_prompt_version)
        continue_chain = continue_prompt.as_langchain() | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        if self.iteration > 1:
            if_loop_prompt: Prompt = super(AgentTemplate,
                                           self).process_prompt(
                agent_input, prompt_version=self.if_loop_prompt_version)
            if_loop_prompt_chain = if_loop_prompt.as_langchain() | llm.as_langchain_runnable(
                self.agent_model.llm_params()) | StrOutputParser()

        for i in range(self.iteration):
            continue_res = await self.async_invoke_chain(continue_chain, agent_input,
                                             input_object, **kwargs)
            res = res + '\n' + continue_res

            if i == self.iteration - 1:
                break
            # judge whether loop
            if not self.if_loop_prompt_version:
                continue
            loop_res = await self.async_invoke_chain(if_loop_prompt_chain, agent_input,
                                         input_object, **kwargs)
            if not self.if_loop(loop_res):
                break

            lc_prompt_template = continue_prompt.as_langchain()
            prompt_input_dict = {key: agent_input[key] for key in
                                 lc_prompt_template.input_variables
                                 if key in agent_input}
            conversation_history.append({
                'user': lc_prompt_template.format(**prompt_input_dict),
                'assistant': continue_res
            })
            agent_input['chat_history'] = json.dumps(conversation_history,
                                                     ensure_ascii=False)

        assemble_memory_output(memory=memory,
                               agent_input=agent_input,
                               content=f"Human: {agent_input.get('input')}, AI: {res}")
        return {**agent_input, 'output': res}

    def if_loop(self, loop_res: str):
        return 'yes' in loop_res

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'RagAgentTemplate':
        super().initialize_by_component_configer(component_configer)
        if hasattr(component_configer, "iteration"):
            self.iteration = component_configer.iteration
        if self.iteration > 0:
            try:
                self.continue_prompt_version = component_configer.continue_prompt_version
            except AttributeError as e:
                LOGGER.error("Contextual iteration agent need continue_prompt_version while iteration > 0")
        if hasattr(component_configer, "if_loop_prompt_version"):
            self.if_loop_prompt_version = component_configer.if_loop_prompt_version
        return self


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: simple_math_tool.py


from agentuniverse.agent.action.tool.tool import Tool, ToolInput


class AddTool(Tool):
    def execute(self, tool_input: ToolInput):
        input_params = tool_input.get_data('input')
        a, b = input_params.split(',')
        result = float(a) + float(b)
        return result


class SubtractTool(Tool):
    def execute(self, tool_input: ToolInput):
        input_params = tool_input.get_data('input')
        a, b = input_params.split(',')
        result = float(a) - float(b)
        return result


class MultiplyTool(Tool):
    def execute(self, tool_input: ToolInput):
        input_params = tool_input.get_data('input')
        a, b = input_params.split(',')
        result = float(a) * float(b)
        return result


class DivideTool(Tool):
    def execute(self, tool_input: ToolInput):
        input_params = tool_input.get_data('input')
        a, b = input_params.split(',')
        result = float(a) / float(b)
        return result



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/8 11:45
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: openai_official_llm_channel.py
from typing import Optional

from agentuniverse.llm.llm_channel.llm_channel import LLMChannel

OPENAI_MAX_CONTEXT_LENGTH = {
    "gpt-3.5-turbo": 4096,
    "gpt-3.5-turbo-0301": 4096,
    "gpt-3.5-turbo-0613": 4096,
    "gpt-3.5-turbo-16k": 16384,
    "gpt-3.5-turbo-16k-0613": 16384,
    "gpt-35-turbo": 4096,
    "gpt-35-turbo-16k": 16384,
    "gpt-3.5-turbo-1106": 16384,
    "gpt-3.5-turbo-0125": 16384,
    "gpt-4-0314": 8192,
    "gpt-4": 8192,
    "gpt-4-32k": 32768,
    "gpt-4-32k-0613": 32768,
    "gpt-4-0613": 8192,
    "gpt-4-1106-preview": 128000,
    "gpt-4-turbo": 128000,
    "gpt-4o": 128000,
    "gpt-4o-2024-05-13": 128000,
}


class OpenAIOfficialLLMChannel(LLMChannel):
    channel_api_base: Optional[str] = "https://api.openai.com/v1"

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return OPENAI_MAX_CONTEXT_LENGTH.get(self.channel_model_name, 128000)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/25 14:58
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: rag_planner.py
"""Rag planner module."""
from langchain_core.output_parsers import StrOutputParser

from agentuniverse.agent.agent_model import AgentModel
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.plan.planner.planner import Planner
from agentuniverse.base.util.agent_util import assemble_memory_input, assemble_memory_output
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.llm.llm import LLM
from agentuniverse.prompt.chat_prompt import ChatPrompt
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.prompt.prompt_model import AgentPromptModel


class RagPlanner(Planner):
    """Rag planner class."""

    def invoke(self, agent_model: AgentModel, planner_input: dict,
               input_object: InputObject) -> dict:
        """Invoke the planner.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
            input_object (InputObject): The input parameters passed by the user.
        Returns:
            dict: The planner result.
        """
        memory: Memory = self.handle_memory(agent_model, planner_input)

        self.run_all_actions(agent_model, planner_input, input_object)

        llm: LLM = self.handle_llm(agent_model)

        prompt: ChatPrompt = self.handle_prompt(agent_model, planner_input)
        process_llm_token(llm, prompt.as_langchain(), agent_model.profile, planner_input)

        assemble_memory_input(memory, planner_input)

        chain = prompt.as_langchain() | llm.as_langchain_runnable(agent_model.llm_params()) | StrOutputParser()

        res = self.invoke_chain(agent_model, chain, planner_input, None, input_object)

        assemble_memory_output(memory=memory,
                               agent_input=planner_input,
                               content=f"Human: {planner_input.get(self.input_key)}, AI: {res}")
        return {**planner_input, self.output_key: res}

    def handle_prompt(self, agent_model: AgentModel, planner_input: dict) -> ChatPrompt:
        """Prompt module processing.

        Args:
            agent_model (AgentModel): Agent model object.
            planner_input (dict): Planner input object.
        Returns:
            ChatPrompt: The chat prompt instance.
        """
        profile: dict = agent_model.profile

        profile_prompt_model: AgentPromptModel = AgentPromptModel(introduction=profile.get('introduction'),
                                                                  target=profile.get('target'),
                                                                  instruction=profile.get('instruction'))

        # get the prompt by the prompt version
        prompt_version: str = profile.get('prompt_version')
        version_prompt: Prompt = PromptManager().get_instance_obj(prompt_version)

        if version_prompt is None and not profile_prompt_model:
            raise Exception("Either the `prompt_version` or `introduction & target & instruction`"
                            " in agent profile configuration should be provided.")
        if version_prompt:
            version_prompt_model: AgentPromptModel = AgentPromptModel(
                introduction=getattr(version_prompt, 'introduction', ''),
                target=getattr(version_prompt, 'target', ''),
                instruction=getattr(version_prompt, 'instruction', ''))
            profile_prompt_model = profile_prompt_model + version_prompt_model

        chat_prompt = ChatPrompt().build_prompt(profile_prompt_model, self.prompt_assemble_order)
        image_urls: list = planner_input.pop('image_urls', []) or []
        if image_urls:
            chat_prompt.generate_image_prompt(image_urls)
        return chat_prompt



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: wikipedia_query.py


from langchain_community.tools import WikipediaQueryRun
from langchain_community.utilities import WikipediaAPIWrapper

from agentuniverse.agent.action.tool.common_tool.langchain_tool import LangChainTool


class WikipediaTool(LangChainTool):
    def init_langchain_tool(self, component_configer):
        wrapper = WikipediaAPIWrapper()
        return WikipediaQueryRun(api_wrapper=wrapper)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: request_tool.py


from typing import Optional

from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger
from agentuniverse.base.util.logging.logging_util import LOGGER
from langchain_community.utilities.requests import GenericRequestsWrapper
from langchain_core.utils.json import parse_json_markdown


class RequestTool(Tool):
    method: Optional[str] = 'GET'
    headers: Optional[dict] = {}
    response_content_type: Optional[str] = 'text'
    requests_wrapper: Optional[GenericRequestsWrapper] = None
    json_parser: Optional[bool] = False

    @staticmethod
    def _clean_url(url: str) -> str:
        """Strips quotes from the url."""
        return url.strip("\"'")

    def execute(self, tool_input: ToolInput):
        input_params: str = tool_input.get_data('input')
        if self.json_parser:
            try:
                parse_data = parse_json_markdown(input_params)
                return self.execute_by_method(**parse_data)
            except Exception as e:
                LOGGER.error(f'execute request error input{input_params} error{e}')
                return str(e)
        else:
            return self.execute_by_method(input_params)

    def execute_by_method(self, url: str, data: dict = None, **kwargs):
        url = self._clean_url(url)
        if self.method == 'GET':
            return self.requests_wrapper.get(url)
        elif self.method == 'POST':
            return self.requests_wrapper.post(url, data=data)
        elif self.method == 'PUT':
            return self.requests_wrapper.put(url, data=data)
        elif self.method == 'DELETE':
            return self.requests_wrapper.delete(url)
        else:
            raise ValueError(f"Unsupported method: {self.method}")

    def initialize_by_component_configer(self, component_configer: ToolConfiger) -> 'Tool':
        """
        :param component_configer:
        :return:
        """
        self.headers = component_configer.configer.value.get('headers')
        self.method = component_configer.configer.value.get('method')
        self.response_content_type = component_configer.configer.value.get('response_content_type')
        if 'json_parser' in component_configer.configer.value:
            self.json_parser = component_configer.configer.value.get('json_parser')
        self.requests_wrapper = GenericRequestsWrapper(
            headers=self.headers,
            response_content_type=self.response_content_type
        )
        return super().initialize_by_component_configer(component_configer)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time : 2025/2/12 14:37
# @Author : wozhapen
# @mail : wozhapen@gmail.com
# @FileName :gemini_openai_style_llm.py

from typing import Optional, Any, Union, Iterator, AsyncIterator

from pydantic import Field

from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.llm.openai_style_llm import OpenAIStyleLLM

# Define maximum context lengths for different Gemini models
GEMINI_MAX_CONTEXT_LENGTH = {
    "gemini-2.0-flash": 1048576  # Example value, adjust as needed
    # Add more Gemini models and their context lengths as needed
}


class GeminiOpenAIStyleLLM(OpenAIStyleLLM):
    """
        Gemini OpenAI style LLM
        Args:
            api_key: API key for the model ,from Google AI Studio or Vertex AI
            api_base: API base URL for the model (optional, depending on the service)
    """

    api_key: Optional[str] = Field(default_factory=lambda: get_from_env("GOOGLE_API_KEY"))
    api_base: Optional[str] = Field(default_factory=lambda: get_from_env("GOOGLE_API_BASE"))
    proxy: Optional[str] = Field(default_factory=lambda: get_from_env("GOOGLE_PROXY"))
    organization: Optional[str] = None  # Gemini does not require organization.

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """ The call method of the LLM.

        Users can customize how the model interacts by overriding call method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return super()._call(messages, **kwargs)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """ The async call method of the LLM.

        Users can customize how the model interacts by overriding acall method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return await super()._acall(messages, **kwargs)

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return GEMINI_MAX_CONTEXT_LENGTH.get(self.model_name, 8000)  # Default context length if model not found

    """ 
        The current Google client does not support setting a proxy, 
        therefore local debugging is not supported.
    """
    # def get_num_tokens(self, text: str) -> int:
    #     """
    #       use genai
    #     """
    #     try:
    #         from google import genai
    #         client = genai.Client(api_key=self.api_key)
    #         response = client.models.count_tokens(
    #             model=self.model_name,
    #             contents=text,
    #         )
    #         return response.total_tokens
    #     except Exception as e:
    #         print(
    #             f"Warning: Unable to accurately count tokens for Gemini. Error: {e}")
    #         # Fallback:  Simple word count as a very rough estimate
    #         return len(text.split())



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 13:52
# @Author  : weizjajj
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: langchain_openai_style_instance.py

from typing import Any, List, Optional, AsyncIterator, Iterator, Mapping, Dict, Type, Union

from langchain.callbacks.manager import AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun
from langchain.schema import BaseMessage, ChatResult
from langchain_community.chat_models.openai import _create_retry_decorator
from langchain_community.utils.openai import is_openai_v1
from langchain_core.language_models.chat_models import generate_from_stream, agenerate_from_stream
from langchain_core.messages import AIMessageChunk, get_buffer_string, BaseMessageChunk, HumanMessageChunk, \
    SystemMessageChunk, FunctionMessageChunk, ToolMessageChunk, ChatMessageChunk, HumanMessage, AIMessage, \
    SystemMessage, FunctionMessage, ToolMessage, ChatMessage
from langchain_core.outputs import ChatGenerationChunk, ChatGeneration
from langchain_community.chat_models import ChatOpenAI
from pydantic.v1 import BaseModel

from agentuniverse.llm.llm import LLM


async def acompletion_with_retry(
        llm: ChatOpenAI,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
) -> Any:
    """Use tenacity to retry the async completion call."""
    if is_openai_v1():
        return await llm.llm.acall(**kwargs)

    retry_decorator = _create_retry_decorator(llm, run_manager=run_manager)

    @retry_decorator
    async def _completion_with_retry(**kwargs: Any) -> Any:
        # Use OpenAI's async api https://github.com/openai/openai-python#async-api
        return await llm.llm.acall(**kwargs)

    return await _completion_with_retry(**kwargs)


def _convert_delta_to_message_chunk(
        _dict: Mapping[str, Any], default_class: Type[BaseMessageChunk]
) -> BaseMessageChunk:
    role = _dict.get("role")
    content = _dict.get("content") or ""
    additional_kwargs: Dict = {}
    if _dict.get("function_call"):
        function_call = dict(_dict["function_call"])
        if "name" in function_call and function_call["name"] is None:
            function_call["name"] = ""
        additional_kwargs["function_call"] = function_call
    if _dict.get("tool_calls"):
        additional_kwargs["tool_calls"] = _dict["tool_calls"]

    if _dict.get("reasoning_content"):
        additional_kwargs["reasoning_content"] = _dict["reasoning_content"]

    if role == "user" or default_class == HumanMessageChunk:
        return HumanMessageChunk(content=content)
    elif role == "assistant" or default_class == AIMessageChunk:
        return AIMessageChunk(content=content, additional_kwargs=additional_kwargs)
    elif role == "system" or default_class == SystemMessageChunk:
        return SystemMessageChunk(content=content)
    elif role == "function" or default_class == FunctionMessageChunk:
        return FunctionMessageChunk(content=content, name=_dict["name"])
    elif role == "tool" or default_class == ToolMessageChunk:
        return ToolMessageChunk(content=content, tool_call_id=_dict["tool_call_id"])
    elif role or default_class == ChatMessageChunk:
        return ChatMessageChunk(content=content, role=role)
    else:
        return default_class(content=content)


def convert_dict_to_message(_dict: Mapping[str, Any]) -> BaseMessage:
    """Convert a dictionary to a LangChain message.

    Args:
        _dict: The dictionary.

    Returns:
        The LangChain message.
    """
    role = _dict.get("role")
    if role == "user":
        return HumanMessage(content=_dict.get("content", ""))
    elif role == "assistant":
        # Fix for azure
        # Also OpenAI returns None for tool invocations
        content = _dict.get("content", "") or ""
        additional_kwargs: Dict = {}
        if function_call := _dict.get("function_call"):
            additional_kwargs["function_call"] = dict(function_call)
        if tool_calls := _dict.get("tool_calls"):
            additional_kwargs["tool_calls"] = tool_calls
        if reasoning_content := _dict.get("reasoning_content"):
            additional_kwargs["reasoning_content"] = reasoning_content
        return AIMessage(content=content, additional_kwargs=additional_kwargs)
    elif role == "system":
        return SystemMessage(content=_dict.get("content", ""))
    elif role == "function":
        return FunctionMessage(content=_dict.get("content", ""), name=_dict.get("name"))
    elif role == "tool":
        additional_kwargs = {}
        if "name" in _dict:
            additional_kwargs["name"] = _dict["name"]
        return ToolMessage(
            content=_dict.get("content", ""),
            tool_call_id=_dict.get("tool_call_id"),
            additional_kwargs=additional_kwargs,
        )
    else:
        return ChatMessage(content=_dict.get("content", ""), role=role)


class LangchainOpenAIStyleInstance(ChatOpenAI):
    """Langchain OpenAI LLM wrapper."""

    llm: Optional[LLM] = None

    def __init__(self, llm: LLM):
        """The __init__ method.

        The agentUniverse LLM instance is passed to this class as an argument.
        Convert the attributes of agentUniverse(aU) LLM instance to the LangchainOpenAI object for initialization

        Args:
            llm (LLM): the agentUniverse(aU) LLM instance.
        """
        init_params = dict()
        init_params['model_name'] = llm.model_name if llm.model_name else 'gpt-3.5-turbo'
        init_params['temperature'] = llm.temperature if llm.temperature else 0.7
        init_params['request_timeout'] = llm.request_timeout
        init_params['max_tokens'] = llm.max_tokens
        init_params['max_retries'] = llm.max_retries if llm.max_retries else 2
        init_params['streaming'] = llm.streaming if llm.streaming else False
        init_params['openai_api_key'] = llm.api_key if llm.api_key else 'blank'
        init_params['openai_organization'] = llm.organization
        init_params['openai_api_base'] = llm.api_base
        init_params['openai_proxy'] = llm.proxy
        super().__init__(**init_params)
        self.llm = llm

    def _generate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            stream: Optional[bool] = None,
            **kwargs,
    ) -> ChatResult:
        """Run the Langchain OpenAI LLM."""
        should_stream = stream if stream is not None else self.streaming
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        llm_output = self.llm.call(messages=message_dicts, **params)
        if not should_stream:
            return self._create_chat_result(llm_output.raw)
        stream_iter = self.as_langchain_chunk(llm_output)
        return generate_from_stream(stream_iter)

    async def _agenerate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            stream: Optional[bool] = None,
            **kwargs: Any,
    ) -> ChatResult:

        """Asynchronously run the Langchain OpenAI LLM."""
        should_stream = stream if stream is not None else self.streaming
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        llm_output = await self.llm.acall(messages=message_dicts, **params)
        if not should_stream:
            return self._create_chat_result(llm_output.raw)
        stream_iter = self.as_langchain_achunk(llm_output)
        return await agenerate_from_stream(stream_iter)

    @staticmethod
    def as_langchain_chunk(stream, run_manager=None):
        default_chunk_class = AIMessageChunk
        for llm_result in stream:
            chunk = llm_result.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            chunk = ChatGenerationChunk(message=chunk, generation_info=generation_info)
            yield chunk
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)

    @staticmethod
    async def as_langchain_achunk(stream_iterator: AsyncIterator, run_manager=None) \
            -> AsyncIterator[ChatGenerationChunk]:
        default_chunk_class = AIMessageChunk
        async for llm_result in stream_iterator:
            chunk = llm_result.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            chunk = ChatGenerationChunk(message=chunk, generation_info=generation_info)
            yield chunk
            if run_manager:
                await run_manager.on_llm_new_token(token=chunk.text, chunk=chunk)

    def get_num_tokens_from_messages(self, messages: List[BaseMessage]) -> int:
        messages_str = get_buffer_string(messages)
        return self.llm.get_num_tokens(messages_str)

    def completion_with_retry(
            self, run_manager: Optional[CallbackManagerForLLMRun] = None, **kwargs: Any
    ) -> Any:
        """Use tenacity to retry the completion call."""
        if is_openai_v1():
            return self.llm.call(**kwargs)

        retry_decorator = _create_retry_decorator(self, run_manager=run_manager)

        @retry_decorator
        def _completion_with_retry(**kwargs: Any) -> Any:
            return self.llm.call(**kwargs)

        return _completion_with_retry(**kwargs)

    def _stream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        for chunk in self.completion_with_retry(
                messages=message_dicts, run_manager=run_manager, **params
        ):
            chunk = chunk.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            cg_chunk = ChatGenerationChunk(
                message=chunk, generation_info=generation_info
            )
            if run_manager:
                run_manager.on_llm_new_token(cg_chunk.text, chunk=cg_chunk)
            yield cg_chunk

    async def _astream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        async for chunk in await acompletion_with_retry(
                self, messages=message_dicts, run_manager=run_manager, **params
        ):
            chunk = chunk.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = _convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            cg_chunk = ChatGenerationChunk(
                message=chunk, generation_info=generation_info
            )
            if run_manager:
                await run_manager.on_llm_new_token(token=cg_chunk.text, chunk=cg_chunk)
            yield cg_chunk

    def _create_chat_result(self, response: Union[dict, BaseModel]) -> ChatResult:
        generations = []
        if not isinstance(response, dict):
            response = response.dict()
        for res in response["choices"]:
            message = convert_dict_to_message(res["message"])
            generation_info = dict(finish_reason=res.get("finish_reason"))
            if "logprobs" in res:
                generation_info["logprobs"] = res["logprobs"]
            gen = ChatGeneration(
                message=message,
                generation_info=generation_info,
            )
            generations.append(gen)
        token_usage = response.get("usage", {})
        llm_output = {
            "token_usage": token_usage,
            "model_name": self.model_name,
            "system_fingerprint": response.get("system_fingerprint", ""),
        }
        return ChatResult(generations=generations, llm_output=llm_output)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/1/17 10:40
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: llm_input_log_sink.py

from typing import Union

from agentuniverse.base.util.logging.log_sink.base_file_log_sink import BaseFileLogSink
from agentuniverse.base.util.logging.log_type_enum import LogTypeEnum
from agentuniverse.base.util.monitor.monitor import Monitor


class LLMInputLogSink(BaseFileLogSink):
    log_type: LogTypeEnum = LogTypeEnum.llm_input

    def process_record(self, record):
        record["message"] = self.generate_log(
            llm_input=record['extra'].get('llm_input')
        )

    def generate_log(self, llm_input: Union[str, dict]) -> str:
        return Monitor.get_invocation_chain_str() + f" LLM get an input."



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/22 16:31
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: neo4j_store.py

import json
from typing import List, Any, Optional
try:
    from neo4j import GraphDatabase, AsyncGraphDatabase
    import pandas as pd
except ImportError:
    raise ImportError(
        "The functionality you are trying to use requires the neo4j and pandas package. "
        "You can install it by running: pip install neo4j pandas"
    )


from agentuniverse.agent.action.knowledge.store.graph_document import GraphDocument
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.agent.action.knowledge.store.store import Store
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class Neo4jStore(Store):

    uri: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None
    database: Optional[str] = None
    driver: Any = None
    async_driver: Any = None

    def _new_client(self) -> Any:
        if self.database:
            self.driver = GraphDatabase.driver(self.uri, auth=(self.user, self.password), database=self.database)
        else:
            self.driver = GraphDatabase.driver(self.uri,
                                               auth=(self.user, self.password))

    def _new_async_client(self) -> Any:
        if self.database:
            self.async_driver = AsyncGraphDatabase.driver(self.uri,
                                               auth=(self.user, self.password),
                                               database=self.database)
        else:
            self.async_driver = AsyncGraphDatabase.driver(self.uri,
                                               auth=(self.user, self.password))

    def execute_cypher(self, query_str, param=None, return_data=True):
        df_result = self._execute_cypher(self.driver.session(), query_str, param, return_data)
        return df_result

    @staticmethod
    def _execute_cypher(session, query_str, param=None, return_data=True):
        df_result = pd.DataFrame()
        if param is None:
            result = session.run(query_str)
        else:
            result = session.run(query_str, **param)

        if return_data:
            result_list = []
            for resulti in result:
                result_list.append(dict(resulti))

            df_result = pd.DataFrame(result_list)
        session.close()
        return df_result


    def query(self, query: Query, **kwargs) -> List[Document]:
        query_type = query.ext_info.get("query_type", "")

        if query_type == "direct_cypher":
            cypher_query = query.query_str
            query_params = query.ext_info.get("query_params", {})
            records = self.execute_cypher(cypher_query, query_params)
            return self._records_to_documents(cypher_query, records)

        elif query_type == "llm_generate_cypher":
            schema_info = self._get_graph_schema_info()

            llm_cypher = self._llm_generate_cypher(
                query.query_str,
                schema_info
            )
            query_params = query.ext_info.get("query_params", {})
            records = self.execute_cypher(llm_cypher, query_params)
            return self._records_to_documents(query.query_str, records)

        elif query_type == "node_ids_query":
            node_ids = query.query_str
            if not node_ids:
                return []
            cypher_query = self._build_node_ids_query(json.loads(query.query_str))
            query_params = query.ext_info.get("query_params", {})
            records = self.execute_cypher(cypher_query, query_params)
            return self._records_to_documents(node_ids, records)
        else:
            raise NotImplementedError('This query type is not allowed in neo4j store.')

    def _records_to_documents(self, text, records: pd.DataFrame) -> List[Document]:
        documents = [
            GraphDocument(
                text=text,
                graph_data=records
            )
        ]

        return documents

    def _get_graph_schema_info(self) -> dict:
        cypher = "CALL apoc.meta.schema() YIELD value RETURN value"
        session = self.driver.session()
        try:
            result = session.run(cypher)
            schema = [record["value"] for record in result]
            if schema:
                return schema[0]
            return {}
        finally:
            session.close()

    def _llm_generate_cypher(self, natural_language_query: str,
                                      schema_info: dict) -> str:
        return "MATCH (n) RETURN n LIMIT 10"


    def _build_node_ids_query(self, node_ids: List[int]) -> str:
        ids_str = ", ".join(str(i) for i in node_ids)
        return f"MATCH (n) WHERE id(n) IN [{ids_str}] RETURN n"


    def insert_document(self, documents: List[Document], **kwargs: Any):
        session = self.driver.session()
        try:
            for doc in documents:
                cypher = f"""
                MERGE (d:Document {{id: '{doc.id}'}})
                SET d.text = $text,
                    d.metadata = $metadata
                """
                session.run(cypher, text=doc.text, metadata=doc.metadata)
        finally:
            session.close()


    def upsert_document(self, documents: List[Document], **kwargs):

        self.insert_document(documents, **kwargs)

    def update_document(self, documents: List[Document], **kwargs):
        session = self.driver.session()
        try:
            for doc in documents:
                cypher = f"""
                MATCH (d:Document {{id: '{doc.id}'}})
                SET d.text = $text,
                    d.metadata = $metadata
                """
                session.run(cypher, text=doc.text, metadata=doc.metadata)
        finally:
            session.close()

    def _initialize_by_component_configer(self,
                                          neo4j_store_configer: ComponentConfiger) -> 'Neo4jStore':
        super()._initialize_by_component_configer(neo4j_store_configer)
        if hasattr(neo4j_store_configer, "uri"):
            self.uri = neo4j_store_configer.uri
        if hasattr(neo4j_store_configer, "user"):
            self.user = neo4j_store_configer.user
        if hasattr(neo4j_store_configer, "password"):
            self.password = neo4j_store_configer.password
        if hasattr(neo4j_store_configer, "database"):
            self.database = neo4j_store_configer.database
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/7 19:26
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: ollama_llm_channel.py
import json
from typing import Optional, Union, Iterator, AsyncIterator

from langchain_core.language_models import BaseLanguageModel
from ollama import Options

from agentuniverse.agent.memory.message import Message
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.llm.llm_channel.langchain_instance.ollama_channel_langchain_instance import \
    OllamaChannelLangchainInstance
from agentuniverse.llm.llm_channel.llm_channel import LLMChannel
from agentuniverse.llm.llm_output import LLMOutput


class OllamaLLMChannel(LLMChannel):
    channel_api_base: Optional[str] = "http://localhost:11434"

    def _initialize_by_component_configer(self, component_configer: ComponentConfiger) -> 'OllamaLLMChannel':
        super()._initialize_by_component_configer(component_configer)
        return self

    def as_langchain(self) -> BaseLanguageModel:
        return OllamaChannelLangchainInstance(self)

    def _new_client(self):
        if self.client:
            return self.client
        from ollama import Client
        return Client(
            host=self.channel_api_base,
        )

    def _new_async_client(self):
        if self.async_client:
            return self.async_client
        from ollama import AsyncClient
        return AsyncClient(
            host=self.channel_api_base,
        )

    def _options(self):
        return Options(**{
            "num_ctx": self.max_context_length(),
            "num_predict": self.max_tokens,
            "temperature": self.temperature,
            "timeout": self.request_timeout,
            **(self.ext_info if self.ext_info else {}),
        })

    def _call(self, messages, stop=None, **kwargs) -> Union[LLMOutput, Iterator[LLMOutput]]:
        should_stream = kwargs.pop("stream", self.streaming)
        client = self._new_client()
        options = self._options()
        options.setdefault("stop", stop)
        res = client.chat(model=self.channel_model_name, messages=messages, options=options, stream=should_stream)
        if should_stream:
            return self.generate_result(res)
        else:
            return LLMOutput(text=res.get("message").get('content'), raw=json.dumps(res),
                             message=Message.from_dict(res.get("message")))

    async def _acall(self, messages, stop=None, **kwargs) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        client = self._new_async_client()
        should_stream = kwargs.pop("stream", self.streaming)
        options = self._options()
        options.setdefault("stop", stop)
        res = await client.chat(model=self.channel_model_name, messages=messages, options=options, stream=should_stream)
        if not should_stream:
            return LLMOutput(text=res.get("message").get('content'), raw=json.dumps(res),
                             message=Message.from_dict(res.get("message")))
        if should_stream:
            return self.agenerate_result(res)

    def generate_result(self, data):
        for line in data:
            yield LLMOutput(text=line.get("message").get('content'), raw=json.dumps(line),
                            message=Message.from_dict(line.get("message")))

    async def agenerate_result(self, data):
        async for line in data:
            yield LLMOutput(text=line.get("message").get('content'), raw=json.dumps(line),
                            message=Message.from_dict(line.get("message")))



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/2/28 19:49
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: default_llm_configer.py
from typing import Optional

from agentuniverse.base.annotation.singleton import singleton
from ..configer import Configer


@singleton
class DefaultLLMConfiger(Configer):
    """
    A singleton class responsible for managing the default LLM configuration
    used in the agent system.

    This class reads and parses a TOML configuration file to determine the default LLM instance
    that will be used in the agent when no specific model is manually designated.
    """
    default_llm: Optional[str] = None

    def __init__(self, config_path: str = None):
        super().__init__(config_path)
        if config_path:
            try:
                self.load()
            except FileNotFoundError as e:
                print(f"Configuration file not found at path: {config_path}. "
                      f"The default LLM configuration will not be loaded. "
                      f"Error is {str(e)}")
        if self.value:
            self.default_llm = self.value.get('DEFAULT', {}).get('default_llm', None)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/5 14:37
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: text_splitter.py
from typing import List, Optional
from langchain.text_splitter import CharacterTextSplitter as Splitter

from agentuniverse.agent.action.knowledge.doc_processor.doc_processor import \
    DocProcessor
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class CharacterTextSplitter(DocProcessor):
    """Character-based text splitter for document processing.
    
    This class splits documents into smaller chunks based on character separators,
    with configurable chunk size and overlap parameters.
    
    Attributes:
        chunk_size: The size of each text chunk.
        chunk_overlap: The number of characters to overlap between chunks.
        separator: The character sequence used to split text.
        splitter: The underlying LangChain text splitter instance.
    """
    chunk_size: int = 200
    chunk_overlap: int = 20
    separator: str = "/n/n"
    __splitter: Optional[Splitter] = None


    @property
    def splitter(self) -> Splitter:
        if not self.__splitter:
            self.__splitter = Splitter(separator=self.separator,
                                 chunk_size=self.chunk_size,
                                 chunk_overlap=self.chunk_overlap)
        return self.__splitter


    def _process_docs(self, origin_docs: List[Document], query: Query = None) -> \
            List[Document]:
        """Process documents by splitting them into smaller chunks.
        
        Args:
            origin_docs: List of original documents to be processed.
            query: Optional query object that may influence the processing.
            
        Returns:
            List[Document]: List of processed document chunks.
        """
        lc_doc_list = self.splitter.split_documents(Document.as_langchain_list(
            origin_docs
        ))
        return Document.from_langchain_list(lc_doc_list)

    def _initialize_by_component_configer(self,
                                         doc_processor_configer: ComponentConfiger) -> 'DocProcessor':
        """Initialize the splitter using configuration from a ComponentConfiger.
        
        Args:
            doc_processor_configer: Configuration object containing splitter parameters.
            
        Returns:
            DocProcessor: The initialized document processor instance.
        """
        super()._initialize_by_component_configer(doc_processor_configer)
        if hasattr(doc_processor_configer, "chunk_size"):
            self.chunk_size = doc_processor_configer.chunk_size
        if hasattr(doc_processor_configer, "chunk_overlap"):
            self.chunk_overlap = doc_processor_configer.chunk_overlap
        if hasattr(doc_processor_configer, "separator"):
            self.separator = doc_processor_configer.separator
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/4 15:05
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: __init__.py

from agentuniverse.agent.action.knowledge.doc_processor.types.ast_types import *
from agentuniverse.agent.action.knowledge.doc_processor.types.code_types import *
from agentuniverse.agent.action.knowledge.doc_processor.types.metrics_types import *



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/1/16 16:23
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: agent_output_log_sink.py
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.base.util.logging.log_sink.base_file_log_sink import BaseFileLogSink
from agentuniverse.base.util.logging.log_type_enum import LogTypeEnum
from agentuniverse.base.util.monitor.monitor import Monitor


class AgentInvocationLogSink(BaseFileLogSink):
    log_type: LogTypeEnum = LogTypeEnum.agent_invocation

    def process_record(self, record):
        record["message"] = self.generate_log(
            cost_time=record['extra'].get('cost_time'),
            agent_output=record['extra'].get('agent_output')
        )

    def generate_log(self, cost_time: float, agent_output: OutputObject) -> str:
        log_str = f" Agent cost {cost_time:.2f} seconds"
        return Monitor.get_invocation_chain_str() + log_str + f" Agent output is {agent_output.to_json_str()}"



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/13 13:54
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: tool_manager.py
import importlib

from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger


@singleton
class ToolManager(ComponentManagerBase):
    """The ToolManager class, which is used to manage the tools."""

    def __init__(self):
        """Initialize the ToolManager."""
        super().__init__(ComponentEnum.TOOL)

    def get_instance_obj(self, component_instance_name: str,
                         appname: str = None, new_instance: bool = True) -> Tool:
        """Return the tool instance object."""
        if component_instance_name == "__default_instance__":
            return self.get_default_instance(new_instance)
        appname = appname or ApplicationConfigManager().app_configer.base_info_appname
        instance_code = f'{appname}.{self._component_type.value.lower()}.{component_instance_name}'
        instance_obj = self._instance_obj_map.get(instance_code)
        # If the instance does not exist, try to create it using the configuration
        if instance_obj is None:
            # Retrieve the tool configuration map
            tool_configer_map: dict[str, ToolConfiger] = ApplicationConfigManager().app_configer.tool_configer_map
            if tool_configer_map and component_instance_name in tool_configer_map.keys():
                tool_configer = tool_configer_map.get(component_instance_name)
                if tool_configer:
                    # Dynamically import the module and retrieve the class specified in the configuration
                    if tool_configer.meta_class:
                        metadata_module = '.'.join(tool_configer.meta_class.split('.')[:-1])
                        metadata_class = tool_configer.meta_class.split('.')[-1]
                        module = importlib.import_module(metadata_module)
                        component_clz = getattr(module, metadata_class)
                    else:
                        module = importlib.import_module(tool_configer.metadata_module)
                        component_clz = getattr(module, tool_configer.metadata_class)
                    # Initialize the tool instance using the configuration
                    instance_obj: Tool = component_clz().initialize_by_component_configer(tool_configer)
                    if instance_obj:
                        instance_obj.component_config_path = tool_configer.configer.path
                        self.register(instance_obj.get_instance_code(), instance_obj)
        # Return a new copy of the instance if required, otherwise return the existing instance
        if instance_obj and new_instance:
            return instance_obj.create_copy()
        return instance_obj



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/03/04 14:09
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: code_ast_processor.py
import json
from typing import List, Dict, Any, Optional, cast

from agentuniverse.agent.action.knowledge.doc_processor.types.ast_types import AstNode, AstNodePoint, CodeBoundary
from agentuniverse.agent.action.knowledge.doc_processor.types.code_types import CodeFeatures, CodeRepresentation, ChunkRepresentation
from agentuniverse.agent.action.knowledge.doc_processor.types.metrics_types import CodeMetrics

from agentuniverse.agent.action.knowledge.doc_processor.doc_processor import DocProcessor
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


class CodeAstProcessor(DocProcessor):

    max_depth: int = 8
    language_dir: str = None
    chunk_size: int = 1000
    chunk_overlap: int = 200
    max_node_len: int = 100
    _parser: Optional[Any] = None
    _languages: Dict[str, Any] = {}

    def _process_docs(self, origin_docs: List[Document], query: Query = None) -> List[Document]:
        result_docs = []
        for doc in origin_docs:
            code = doc.text
            language = doc.metadata.get('language', 'unknown') if doc.metadata else 'unknown'
            metadata = doc.metadata.copy() if doc.metadata else {}
            metadata['document_type'] = 'code_ast'
            result_docs.extend(self._process_with_tree_sitter(code, language, metadata))
        return result_docs

    def _process_with_tree_sitter(self, code: str, language: str,
                                  metadata: Dict[str, Any]) -> List[Document]:
        def _ensure_language() -> None:
            if language not in self._languages:
                try:
                    from tree_sitter import Language
                    from importlib import import_module
                    module_name = f'tree_sitter_{language}'
                    lang_module = import_module(module_name)
                    self._languages[language] = Language(lang_module.language())
                except ImportError:
                    raise ImportError(
                        f"Could not import {module_name}. Install with: pip install {module_name}")
        _ensure_language()

        result_docs: List[Document] = []
        self._parser.language = self._languages[language]
        tree = self._parser.parse(bytes(code, "utf8"))
        ast_json: AstNode = self._convert_tree_to_json(tree.root_node, code)
        features: CodeFeatures = self._extract_features(tree.root_node, code, language)
        repr: CodeRepresentation = {
            "ast": ast_json,
            "features": features,
            "language": language,
            "code_length": len(code)
        }
        metadata['processing_method'] = 'tree_sitter'
        ast_doc: Document = Document(
            text=json.dumps(repr),
            metadata=metadata
        )
        result_docs.append(ast_doc)
        if len(code) > self.chunk_size:
            chunk_docs = self._generate_code_chunks(
                code, tree.root_node, language, metadata)
            result_docs.extend(chunk_docs)
        return result_docs

    def _convert_tree_to_json(self, node, code: str, depth: int = 0) -> AstNode:
        if depth > self.max_depth:
            return cast(AstNode, {"type": "max_depth_reached"})
        if not node:
            return cast(AstNode, {})

        start_byte, end_byte = node.start_byte, node.end_byte
        text = code[start_byte:end_byte] if end_byte <= len(code) else ""

        start_point: AstNodePoint = {"row": node.start_point[0], "column": node.start_point[1]}
        end_point: AstNodePoint = {"row": node.end_point[0], "column": node.end_point[1]}

        result: AstNode = {
            "type": node.type,
            "start_point": start_point,
            "end_point": end_point,
            "start_byte": start_byte,
            "end_byte": end_byte
        }

        if len(text) < self.max_node_len or node.child_count == 0:
            result["text"] = text

        if node.child_count > 0:
            children = []
            for child in node.children:
                child_json = self._convert_tree_to_json(child, code, depth + 1)
                if child_json:
                    children.append(child_json)
            result["children"] = children

        return result

    def _extract_features(self, node: Any, code: str, language: str) -> CodeFeatures:

        features: CodeFeatures = {
            "node_counts": self._count_node_types(node),
            "code_metrics": self._calculate_code_metrics(code, language),
            "identifier_count": 0,
            "function_count": 0,
            "class_count": 0,
            "statement_count": 0
        }

        cursor = node.walk()

        def _visit():
            nonlocal features
            current_node = cursor.node

            if current_node.type == "identifier":
                features["identifier_count"] += 1
            elif current_node.type in ("function_definition", "method_definition", "function_declaration"):
                features["function_count"] += 1
            elif current_node.type in ("class_definition", "class_declaration"):
                features["class_count"] += 1
            elif "statement" in current_node.type:
                features["statement_count"] += 1

            if cursor.goto_first_child():
                _visit()
                cursor.goto_parent()
            if cursor.goto_next_sibling():
                _visit()

        _visit()
        return features

    def _count_node_types(self, root_node) -> Dict[str, int]:
        counts = {}

        def _traverse(node):
            if node.type not in counts:
                counts[node.type] = 0
            counts[node.type] += 1

            for child in node.children:
                _traverse(child)

        _traverse(root_node)
        return counts

    def _calculate_code_metrics(self, code: str, language: str) -> CodeMetrics:

        lines = code.splitlines()
        code_lines = [line.strip() for line in lines if line.strip(
        ) and not line.strip().startswith(('#', '//', '/*', '*', '*/'))]

        metrics: CodeMetrics = {
            "line_count": len(lines),
            "code_line_count": len(code_lines),
            "avg_line_length": sum(len(line) for line in code_lines) / max(len(code_lines), 1),
            "max_line_length": max([len(line) for line in code_lines]) if code_lines else 0,
            "character_count": len(code)
        }
        return metrics

    def _generate_code_chunks(
            self,
            code: str,
            root_node,
            language: str,
            metadata: Dict[str, Any]) -> List[Document]:
        chunks = []
        lines = code.splitlines()
        boundaries = []

        def _collect_declarations(node, path=""):

            if node.type in ("function_definition", "method_definition", "class_definition",
                             "function_declaration", "method_declaration", "class_declaration"):
                start_line = node.start_point[0]
                end_line = node.end_point[0]

                if end_line - start_line >= 3:
                    node_type = "function" if "function" in node.type else "class"
                    name = None

                    for child in node.children:
                        if child.type == "identifier":
                            name = code[child.start_byte:child.end_byte]
                            break

                    boundary: CodeBoundary = {
                        "start": start_line,
                        "end": end_line,
                        "type": node_type,
                        "name": name,
                        "node": node
                    }
                    boundaries.append(boundary)

        traverse_cursor = root_node.walk()

        def _traverse_nodes():
            current_node = traverse_cursor.node

            _collect_declarations(current_node)

            if traverse_cursor.goto_first_child():
                _traverse_nodes()
                traverse_cursor.goto_parent()

            if traverse_cursor.goto_next_sibling():
                _traverse_nodes()

        _traverse_nodes()

        if boundaries:
            for boundary in boundaries:
                start_line = boundary["start"]
                end_line = boundary["end"]

                if end_line - start_line < 3:
                    continue

                chunk_code = "\n".join(lines[start_line:end_line + 1])

                chunk_metadata = metadata.copy()
                chunk_metadata["chunk_type"] = boundary["type"]
                chunk_metadata["chunk_name"] = boundary["name"]
                chunk_metadata["start_line"] = start_line
                chunk_metadata["end_line"] = end_line
                chunk_metadata["document_type"] = "code_chunk"

                try:
                    tree = self._parser.parse(bytes(chunk_code, "utf8"))
                    ast_json = self._convert_tree_to_json(tree.root_node, chunk_code)

                    representation: ChunkRepresentation = {
                        "ast": ast_json,
                        "code": chunk_code,
                        "language": language,
                        "name": boundary["name"],
                        "type": boundary["type"]
                    }

                    chunk_doc = Document(
                        text=json.dumps(representation),
                        metadata=chunk_metadata
                    )
                    chunks.append(chunk_doc)
                except BaseException:

                    chunk_metadata["ast_available"] = False
                    chunks.append(Document(text=chunk_code, metadata=chunk_metadata))

        if len(chunks) < 2 and len(lines) > self.chunk_size:
            for i in range(0, len(lines), self.chunk_size - self.chunk_overlap):
                end_idx = min(i + self.chunk_size, len(lines))
                chunk_code = "\n".join(lines[i:end_idx])

                chunk_metadata = metadata.copy()
                chunk_metadata["chunk_type"] = "window"
                chunk_metadata["start_line"] = i
                chunk_metadata["end_line"] = end_idx - 1
                chunk_metadata["document_type"] = "code_chunk"

                chunks.append(Document(text=chunk_code, metadata=chunk_metadata))

                if end_idx >= len(lines):
                    break

        return chunks

    def _initialize_by_component_configer(
            self, doc_processor_configer: ComponentConfiger) -> 'DocProcessor':
        super()._initialize_by_component_configer(doc_processor_configer)
        try:
            from tree_sitter import Parser
            self._parser = Parser()
            self._languages = {}
        except ImportError:
            raise ImportError(
                "tree-sitter not available. Install with: pip install tree-sitter")

        if hasattr(doc_processor_configer, "max_depth"):
            self.max_depth = doc_processor_configer.max_depth

        if not hasattr(doc_processor_configer, "language_dir"):
            raise ValueError(
                "language_dir is required - tree-sitter needs compiled language libraries (.so files) to parse code, download from https://tree-sitter.github.io/tree-sitter/#available-parsers")
        self.language_dir = doc_processor_configer.language_dir

        if hasattr(doc_processor_configer, "chunk_size"):
            self.chunk_size = doc_processor_configer.chunk_size

        if hasattr(doc_processor_configer, "chunk_overlap"):
            self.chunk_overlap = doc_processor_configer.chunk_overlap

        if hasattr(doc_processor_configer, "max_node_text_length"):
            self.max_node_len = doc_processor_configer.max_node_len
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/5 16:22
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: log_type_const.py
from enum import Enum


class LogTypeEnum(str, Enum):
    default = 'default'
    sls = 'sls'
    flask_request = 'flask_request'
    flask_response = 'flask_response'
    agent_input = 'agent_input'
    agent_invocation = 'agent_invocation'
    agent_first_token = 'agent_first_token'
    llm_input = 'llm_input'
    llm_invocation = 'llm_invocation'
    tool_input = 'tool_input'
    tool_invocation = 'tool_invocation'


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/8 14:13
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: baichuan_official_llm_channel.py
from typing import Optional

from agentuniverse.llm.llm_channel.llm_channel import LLMChannel

BAICHUAN_MAX_CONTEXT_LENGTH = {
    "Baichuan2-Turbo": 8000,
    "Baichuan2-Turbo-192k": 192000,
    "Baichuan3-Turbo": 8000,
    "Baichuan3-Turbo-128k": 128000,
    "Baichuan4": 8000
}


class BaichuanOfficialLLMChannel(LLMChannel):
    channel_api_base: Optional[str] = "https://api.baichuan-ai.com/v1"

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return BAICHUAN_MAX_CONTEXT_LENGTH.get(self.channel_model_name, 8000)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/20 16:24
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: monitor.py
import datetime
import json
import os
from typing import Union, Optional
from loguru import logger

from pydantic import BaseModel

from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.config.configer import Configer
from agentuniverse.base.context.framework_context_manager import FrameworkContextManager
from agentuniverse.base.util.logging.general_logger import get_context_prefix
from agentuniverse.base.util.logging.log_type_enum import LogTypeEnum
from agentuniverse.base.util.tracing.au_trace_manager import AuTraceManager

LLM_INVOCATION_SUBDIR = "llm_invocation"
AGENT_INVOCATION_SUBDIR = "agent_invocation"


@singleton
class Monitor(BaseModel):
    dir: Optional[str] = './monitor'
    activate: Optional[bool] = False
    log_activate: Optional[bool] = True

    def __init__(self, configer: Configer = None, **kwargs):
        super().__init__(**kwargs)
        if configer:
            config: dict = configer.value.get('MONITOR', {})
            self.dir = config.get('dir', './monitor')
            self.activate = config.get('activate', False)
            self.log_activate = config.get('log_activate', True)

    def trace_llm_input(self, source: str, llm_input: Union[str, dict]) -> None:
        """Trace the llm input."""
        if self.log_activate:
            logger.bind(
                log_type=LogTypeEnum.llm_input,
                llm_input=llm_input,
                context_prefix=get_context_prefix()
            ).info("Trace llm input.")

    def trace_llm_invocation(self, source: str, llm_input: Union[str, dict], llm_output: Union[str, dict],
                             cost_time: float = None) -> None:
        """Trace the llm invocation and save it to the monitor jsonl file."""
        if self.activate:
            try:
                import jsonlines
            except ImportError:
                raise ImportError(
                    "jsonlines is required to trace llm invocation: `pip install jsonlines`"
                )
            # get the current time
            date = datetime.datetime.now()
            llm_invocation = {
                "source": source,
                "date": date.strftime("%Y-%m-%d %H:%M:%S"),
                "llm_input": llm_input,
                "llm_output": llm_output,
            }
            # files are stored in hours
            filename = f"llm_{date.strftime('%Y-%m-%d-%H')}.jsonl"
            # file path to save
            path_save = os.path.join(str(self._get_or_create_subdir(LLM_INVOCATION_SUBDIR)), filename)

            # write to jsonl
            with jsonlines.open(path_save, 'a') as writer:
                writer.write(llm_invocation)

        if self.log_activate:
            logger.bind(
                log_type=LogTypeEnum.llm_invocation,
                used_token=Monitor.get_token_usage(),
                cost_time=cost_time,
                llm_output=llm_output,
                context_prefix=get_context_prefix()
            ).info("Trace llm input.")

    def trace_llm_token_usage(self, llm_obj: object, llm_input: dict, output_str: str) -> None:
        """ Trace the token usage of the given LLM object.
        Args:
            llm_obj(object): LLM object.
            llm_input(dict): Dictionary of LLM input.
            output_str(str): LLM output.
        """
        trace_id = AuTraceManager().get_trace_id()
        # trace token usage for a complete request chain based on trace id
        if trace_id:
            old_token_usage: dict = FrameworkContextManager().get_context(trace_id + '_token_usage')
            if old_token_usage is not None:
                token_usage: dict = self.get_llm_token_usage(llm_obj, llm_input, output_str)
                if token_usage:
                    Monitor.add_token_usage(token_usage)

    def trace_agent_input(self, source: str, agent_input: Union[str, dict]) -> None:
        """Trace the agent input."""
        if self.log_activate:
            logger.bind(
                log_type=LogTypeEnum.agent_input,
                agent_input=agent_input,
                context_prefix=get_context_prefix()
            ).info("Trace agent input.")

    def trace_agent_invocation(self, source: str, agent_input: Union[str, dict],
                               agent_output: OutputObject, cost_time: float = None) -> None:
        """Trace the agent invocation and save it to the monitor jsonl file."""
        if self.activate:
            try:
                import jsonlines
            except ImportError:
                raise ImportError(
                    "jsonlines is required to trace llm invocation: `pip install jsonlines`"
                )
            # get the current time
            date = datetime.datetime.now()
            agent_invocation = {
                "source": source,
                "date": date.strftime("%Y-%m-%d %H:%M:%S"),
                "agent_input": self.serialize_obj(agent_input),
                "agent_output": self.serialize_obj(agent_output),
            }
            # files are stored in hours
            filename = f"agent_{source}_{date.strftime('%Y-%m-%d-%H')}.jsonl"
            # file path to save
            path_save = os.path.join(str(self._get_or_create_subdir(AGENT_INVOCATION_SUBDIR)), filename)

            # write to jsonl
            with jsonlines.open(path_save, 'a') as writer:
                writer.write(agent_invocation)

        if self.log_activate:
            logger.bind(
                log_type=LogTypeEnum.agent_invocation,
                cost_time=cost_time,
                agent_output=agent_output,
                context_prefix=get_context_prefix()
            ).info("Trace agent invocation.")

    def trace_tool_input(self, source: str, tool_input: Union[str, dict]) -> None:
        """Trace the tool input."""
        if self.log_activate:
            logger.bind(
                log_type=LogTypeEnum.tool_input,
                tool_input=tool_input,
                context_prefix=get_context_prefix()
            ).info("Trace tool input.")

    def trace_tool_invocation(self, source: str, tool_input: Union[str, dict],
                              tool_output: Union[str, dict], cost_time: float = None) -> None:
        """Trace the tool invocation and save it to the monitor jsonl file."""
        if self.log_activate:
            logger.bind(
                log_type=LogTypeEnum.tool_invocation,
                cost_time=cost_time,
                tool_output=tool_output,
                context_prefix=get_context_prefix()
            ).info("Trace tool invocation.")

    @staticmethod
    def init_invocation_chain():
        """Initialize the invocation chain in the framework context."""
        trace_id = AuTraceManager().get_trace_id()
        if FrameworkContextManager().get_context(trace_id + '_invocation_chain') is None:
            FrameworkContextManager().set_context(trace_id + '_invocation_chain', [])

    @staticmethod
    def init_invocation_chain_bak():
        """Initialize the invocation chain bak version in the framework context."""
        trace_id = AuTraceManager().get_trace_id()
        if FrameworkContextManager().get_context(trace_id + '_invocation_chain_bak') is None:
            FrameworkContextManager().set_context(trace_id + '_invocation_chain_bak', [])

    @staticmethod
    def pop_invocation_chain():
        """Pop the last chain node in invocation chain."""
        trace_id = AuTraceManager().get_trace_id()
        if trace_id is not None:
            invocation_chain: list = FrameworkContextManager().get_context(
                trace_id + '_invocation_chain')
            if invocation_chain:
                invocation_chain.pop()
                FrameworkContextManager().set_context(
                    trace_id + '_invocation_chain', invocation_chain)

    @staticmethod
    def clear_invocation_chain():
        """Clear the invocation chain in the framework context."""
        trace_id = AuTraceManager().get_trace_id()
        if trace_id is not None:
            FrameworkContextManager().del_context(trace_id + '_invocation_chain')
            FrameworkContextManager().del_context(trace_id + '_invocation_chain_bak')

    @staticmethod
    def add_invocation_chain(source: dict):
        """Add the source to the invocation chain"""
        trace_id = AuTraceManager().get_trace_id()
        if trace_id is not None:
            invocation_chain = FrameworkContextManager().get_context(trace_id + '_invocation_chain')
            if invocation_chain is not None:
                invocation_chain.append(source)
                FrameworkContextManager().set_context(trace_id + '_invocation_chain', invocation_chain)
            invocation_chain_bak = FrameworkContextManager().get_context(trace_id + '_invocation_chain_bak')
            if invocation_chain_bak is not None:
                invocation_chain_bak.append(source)

    @staticmethod
    def get_trace_id():
        """Get the trace id in the framework context."""
        return AuTraceManager().get_trace_id()

    @staticmethod
    def get_invocation_chain():
        """Get the invocation chain in the framework context."""
        trace_id = AuTraceManager().get_trace_id()
        return FrameworkContextManager().get_context(trace_id + '_invocation_chain', []) if trace_id is not None else []

    @staticmethod
    def get_invocation_chain_bak():
        """Get the invocation chain bak version in the framework context."""
        trace_id = AuTraceManager().get_trace_id()
        return FrameworkContextManager().get_context(trace_id + '_invocation_chain_bak',
                                                     []) if trace_id is not None else []

    @staticmethod
    def init_token_usage():
        """Initialize the token usage in the framework context."""
        trace_id = AuTraceManager().get_trace_id()
        if FrameworkContextManager().get_context(trace_id + '_token_usage') is None:
            FrameworkContextManager().set_context(trace_id + '_token_usage', {})

    @staticmethod
    def add_token_usage(cur_token_usage: dict):
        """Add the token usage to the framework context."""
        if cur_token_usage is None:
            return
        trace_id = AuTraceManager().get_trace_id()
        if trace_id is not None:
            old_token_usage: dict = FrameworkContextManager().get_context(trace_id + '_token_usage')
            if old_token_usage is not None:
                for key, value in cur_token_usage.items():
                    old_token_usage[key] = old_token_usage[key] + value if key in old_token_usage else value
                FrameworkContextManager().set_context(trace_id + '_token_usage', old_token_usage)

    @staticmethod
    def clear_token_usage():
        """Clear the token usage in the framework context."""
        trace_id = AuTraceManager().get_trace_id()
        if trace_id is not None:
            FrameworkContextManager().del_context(trace_id + '_token_usage')
        FrameworkContextManager().del_context('trace_id')

    @staticmethod
    def get_token_usage():
        """Get the token usage in the framework context."""
        trace_id = AuTraceManager().get_trace_id()
        return FrameworkContextManager().get_context(trace_id + '_token_usage', {}) if trace_id is not None else {}

    @staticmethod
    def get_invocation_chain_str() -> str:
        invocation_chain_str = ''
        invocation_chain = Monitor.get_invocation_chain()
        if len(invocation_chain) > 0:
            invocation_chain_str = ' -> '.join(
                [f"source: {d.get('source', '')}, type: {d.get('type', '')}" for
                 d in invocation_chain]
            )
            invocation_chain_str += ' | '
        return invocation_chain_str

    @staticmethod
    def get_llm_token_usage(llm_obj: object, llm_input: dict, output_str: str) -> dict:
        """ Calculate the token usage of the given LLM object.
        Args:
            llm_obj(object): LLM object.
            llm_input(dict): Dictionary of LLM input.
            output_str(str): LLM output.

        Returns:
            dict: Dictionary of token usage including the completion_tokens, prompt_tokens, and total_tokens.
        """
        try:
            if llm_obj is None or llm_input is None:
                return {}
            messages = llm_input.get('kwargs', {}).pop('messages', None)

            input_str = ''
            if messages is not None and isinstance(messages, list):
                for m in messages:
                    if isinstance(m, dict):
                        input_str += str(m.get('role', '')) + '\n'
                        input_str += str(m.get('content', '')) + '\n'

                    elif isinstance(m, object):
                        if hasattr(m, 'role'):
                            role = m.role
                            if role is not None:
                                input_str += str(m.role) + '\n'
                        if hasattr(m, 'content'):
                            content = m.content
                            if content is not None:
                                input_str += str(m.content) + '\n'

            if input_str == '' or output_str == '':
                return {}

            usage = {}
            # the number of input and output tokens is calculated by the llm `get_num_tokens` method.
            if hasattr(llm_obj, 'get_num_tokens'):
                completion_tokens = llm_obj.get_num_tokens(output_str)
                prompt_tokens = llm_obj.get_num_tokens(input_str)
                total_tokens = completion_tokens + prompt_tokens
                usage = {'completion_tokens': completion_tokens, 'prompt_tokens': prompt_tokens,
                         'total_tokens': total_tokens}
            return usage
        except Exception as e:
            return {}

    def _get_or_create_subdir(self, subdir: str) -> str:
        """Get or create a subdirectory if it doesn't exist in the monitor directory."""
        path = os.path.join(self.dir, subdir)
        os.makedirs(path, exist_ok=True)
        return path

    @staticmethod
    def default_serializer(obj):
        """Default serializer for objects."""
        if isinstance(obj, InputObject):
            return obj.to_dict()
        elif isinstance(obj, OutputObject):
            return obj.to_dict()
        elif isinstance(obj, BaseModel):
            try:
                return obj.dict()
            except TypeError:
                raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")
        else:
            raise TypeError(f"Object of type {type(obj).__name__} is not JSON serializable")

    def serialize_obj(self, obj):
        """Serialize an object and filter out non-serializable values."""
        filtered_obj = self.filter_and_serialize(obj)
        return json.loads(json.dumps(filtered_obj, default=self.default_serializer))

    def filter_and_serialize(self, obj):
        """Recursively filter out non-serializable values from an object."""

        def is_json_serializable(value):
            """Check if value is a JSON serializable object."""
            try:
                json.dumps(value, default=self.default_serializer)
                return True
            except (TypeError, OverflowError):
                return False

        def filter_dict(d):
            return {k: v for k, v in d.items() if is_json_serializable(v)}

        def recursive_filter(o):
            if isinstance(o, dict):
                return filter_dict({k: recursive_filter(v) for k, v in o.items()})
            elif isinstance(o, list):
                return [recursive_filter(i) for i in o if is_json_serializable(i)]
            else:
                return o

        return recursive_filter(obj)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/4/2 15:43
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: default_channel_langchain_instance.py
from typing import Any, List, Optional, AsyncIterator, Iterator, Mapping, Dict, Type, Union

from langchain.callbacks.manager import AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun
from langchain.schema import BaseMessage, ChatResult
from langchain_community.chat_models.openai import _create_retry_decorator
from langchain_community.utils.openai import is_openai_v1
from langchain_core.language_models.chat_models import generate_from_stream, agenerate_from_stream
from langchain_core.messages import AIMessageChunk, get_buffer_string, BaseMessageChunk, HumanMessageChunk, \
    SystemMessageChunk, FunctionMessageChunk, ToolMessageChunk, ChatMessageChunk, HumanMessage, AIMessage, \
    SystemMessage, FunctionMessage, ToolMessage, ChatMessage
from langchain_core.outputs import ChatGenerationChunk, ChatGeneration
from langchain_community.chat_models import ChatOpenAI
from pydantic.v1 import BaseModel


class DefaultChannelLangchainInstance(ChatOpenAI):
    llm_channel: Optional[BaseModel] = None

    def __init__(self, llm_channel):
        init_params = dict()
        init_params['model_name'] = llm_channel.channel_model_name
        init_params['temperature'] = llm_channel.temperature if llm_channel.temperature else 0.7
        init_params['request_timeout'] = llm_channel.request_timeout
        init_params['max_tokens'] = llm_channel.max_tokens
        init_params['max_retries'] = llm_channel.max_retries if llm_channel.max_retries else 2
        init_params['streaming'] = llm_channel.streaming if llm_channel.streaming else False
        init_params['openai_api_key'] = llm_channel.channel_api_key if llm_channel.channel_api_key else 'blank'
        init_params['openai_organization'] = llm_channel.channel_organization
        init_params['openai_api_base'] = llm_channel.channel_api_base
        init_params['openai_proxy'] = llm_channel.channel_proxy
        super().__init__(**init_params)
        self.llm_channel = llm_channel

    def _generate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            stream: Optional[bool] = None,
            **kwargs,
    ) -> ChatResult:
        """Run the Langchain OpenAI LLM."""
        should_stream = stream if stream is not None else self.streaming
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        llm_output = self.llm_channel.call(messages=message_dicts, **params)
        if not should_stream:
            return self._create_chat_result(llm_output.raw)
        stream_iter = self.as_langchain_chunk(llm_output)
        return generate_from_stream(stream_iter)

    async def _agenerate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            stream: Optional[bool] = None,
            **kwargs: Any,
    ) -> ChatResult:

        """Asynchronously run the Langchain OpenAI LLM."""
        should_stream = stream if stream is not None else self.streaming
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        llm_output = await self.llm_channel.acall(messages=message_dicts, **params)
        if not should_stream:
            return self._create_chat_result(llm_output.raw)
        stream_iter = self.as_langchain_achunk(llm_output)
        return await agenerate_from_stream(stream_iter)

    def as_langchain_chunk(self, stream, run_manager=None):
        default_chunk_class = AIMessageChunk
        for llm_result in stream:
            chunk = llm_result.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = self._convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            chunk = ChatGenerationChunk(message=chunk, generation_info=generation_info)
            yield chunk
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)

    async def as_langchain_achunk(self, stream_iterator: AsyncIterator, run_manager=None) \
            -> AsyncIterator[ChatGenerationChunk]:
        default_chunk_class = AIMessageChunk
        async for llm_result in stream_iterator:
            chunk = llm_result.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = self._convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            chunk = ChatGenerationChunk(message=chunk, generation_info=generation_info)
            yield chunk
            if run_manager:
                await run_manager.on_llm_new_token(token=chunk.text, chunk=chunk)

    def get_num_tokens_from_messages(self, messages: List[BaseMessage]) -> int:
        messages_str = get_buffer_string(messages)
        return self.llm_channel.get_num_tokens(messages_str)

    def completion_with_retry(
            self, run_manager: Optional[CallbackManagerForLLMRun] = None, **kwargs: Any
    ) -> Any:
        """Use tenacity to retry the completion call."""
        if is_openai_v1():
            return self.llm_channel.call(**kwargs)

        retry_decorator = _create_retry_decorator(self, run_manager=run_manager)

        @retry_decorator
        def _completion_with_retry(**kwargs: Any) -> Any:
            return self.llm_channel.call(**kwargs)

        return _completion_with_retry(**kwargs)

    async def acompletion_with_retry(
            self,
            llm: ChatOpenAI,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> Any:
        """Use tenacity to retry the async completion call."""
        if is_openai_v1():
            return await llm.llm.acall(**kwargs)

        retry_decorator = _create_retry_decorator(llm, run_manager=run_manager)

        @retry_decorator
        async def _completion_with_retry(**kwargs: Any) -> Any:
            # Use OpenAI's async api https://github.com/openai/openai-python#async-api
            return await llm.llm.acall(**kwargs)

        return await _completion_with_retry(**kwargs)

    def _stream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        for chunk in self.completion_with_retry(
                messages=message_dicts, run_manager=run_manager, **params
        ):
            chunk = chunk.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = self._convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            cg_chunk = ChatGenerationChunk(
                message=chunk, generation_info=generation_info
            )
            if run_manager:
                run_manager.on_llm_new_token(cg_chunk.text, chunk=cg_chunk)
            yield cg_chunk

    async def _astream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs, "stream": True}

        default_chunk_class = AIMessageChunk
        async for chunk in await self.acompletion_with_retry(
                self, messages=message_dicts, run_manager=run_manager, **params
        ):
            chunk = chunk.raw
            if not isinstance(chunk, dict):
                chunk = chunk.dict()
            if len(chunk["choices"]) == 0:
                continue
            choice = chunk["choices"][0]
            chunk = self._convert_delta_to_message_chunk(
                choice["delta"], default_chunk_class
            )
            finish_reason = choice.get("finish_reason")
            generation_info = (
                dict(finish_reason=finish_reason) if finish_reason is not None else None
            )
            default_chunk_class = chunk.__class__
            cg_chunk = ChatGenerationChunk(
                message=chunk, generation_info=generation_info
            )
            if run_manager:
                await run_manager.on_llm_new_token(token=cg_chunk.text, chunk=cg_chunk)
            yield cg_chunk

    def _create_chat_result(self, response: Union[dict, BaseModel]) -> ChatResult:
        generations = []
        if not isinstance(response, dict):
            response = response.dict()
        for res in response["choices"]:
            message = self.convert_dict_to_message(res["message"])
            generation_info = dict(finish_reason=res.get("finish_reason"))
            if "logprobs" in res:
                generation_info["logprobs"] = res["logprobs"]
            gen = ChatGeneration(
                message=message,
                generation_info=generation_info,
            )
            generations.append(gen)
        token_usage = response.get("usage", {})
        llm_output = {
            "token_usage": token_usage,
            "model_name": self.model_name,
            "system_fingerprint": response.get("system_fingerprint", ""),
        }
        return ChatResult(generations=generations, llm_output=llm_output)

    @staticmethod
    def _convert_delta_to_message_chunk(
            _dict: Mapping[str, Any], default_class: Type[BaseMessageChunk]
    ) -> BaseMessageChunk:
        role = _dict.get("role")
        content = _dict.get("content") or ""
        additional_kwargs: Dict = {}
        if _dict.get("function_call"):
            function_call = dict(_dict["function_call"])
            if "name" in function_call and function_call["name"] is None:
                function_call["name"] = ""
            additional_kwargs["function_call"] = function_call
        if _dict.get("tool_calls"):
            additional_kwargs["tool_calls"] = _dict["tool_calls"]

        if _dict.get("reasoning_content"):
            additional_kwargs["reasoning_content"] = _dict["reasoning_content"]

        if role == "user" or default_class == HumanMessageChunk:
            return HumanMessageChunk(content=content)
        elif role == "assistant" or default_class == AIMessageChunk:
            return AIMessageChunk(content=content, additional_kwargs=additional_kwargs)
        elif role == "system" or default_class == SystemMessageChunk:
            return SystemMessageChunk(content=content)
        elif role == "function" or default_class == FunctionMessageChunk:
            return FunctionMessageChunk(content=content, name=_dict["name"])
        elif role == "tool" or default_class == ToolMessageChunk:
            return ToolMessageChunk(content=content, tool_call_id=_dict["tool_call_id"])
        elif role or default_class == ChatMessageChunk:
            return ChatMessageChunk(content=content, role=role)
        else:
            return default_class(content=content)

    def convert_dict_to_message(self, _dict: Mapping[str, Any]) -> BaseMessage:
        """Convert a dictionary to a LangChain message.

        Args:
            _dict: The dictionary.

        Returns:
            The LangChain message.
        """
        role = _dict.get("role")
        if role == "user":
            return HumanMessage(content=_dict.get("content", ""))
        elif role == "assistant":
            # Fix for azure
            # Also OpenAI returns None for tool invocations
            content = _dict.get("content", "") or ""
            additional_kwargs: Dict = {}
            if function_call := _dict.get("function_call"):
                additional_kwargs["function_call"] = dict(function_call)
            if tool_calls := _dict.get("tool_calls"):
                additional_kwargs["tool_calls"] = tool_calls
            if reasoning_content := _dict.get("reasoning_content"):
                additional_kwargs["reasoning_content"] = reasoning_content
            return AIMessage(content=content, additional_kwargs=additional_kwargs)
        elif role == "system":
            return SystemMessage(content=_dict.get("content", ""))
        elif role == "function":
            return FunctionMessage(content=_dict.get("content", ""), name=_dict.get("name"))
        elif role == "tool":
            additional_kwargs = {}
            if "name" in _dict:
                additional_kwargs["name"] = _dict["name"]
            return ToolMessage(
                content=_dict.get("content", ""),
                tool_call_id=_dict.get("tool_call_id"),
                additional_kwargs=additional_kwargs,
            )
        else:
            return ChatMessage(content=_dict.get("content", ""), role=role)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/2/22 10:00
# @Author  : wangyapei 
# @FileName: jina_ai_tool.py

from typing import Optional, Dict
import requests
import time
from pydantic import Field
from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.base.util.env_util import get_from_env
import html


jina_read_url = "https://r.jina.ai/"
jina_search_url = "https://s.jina.ai/"
jina_check_fact_url = "https://g.jina.ai/"


class JinaAITool(Tool):
    """The demo jina ai tool.

    Use jina.ai's API for webpage read, search and fact check.
    
    Note: the api key is not required for webpage read, but required for search and check_fact
          api_key can be found in https://jina.ai/.
    
    """

    mode: str = "read"
    timeout: int = 30
    api_key: Optional[str] = Field(default_factory=lambda: get_from_env("JINA_API_KEY"))
    max_read_content_length: int = Field(10000, description="Maximum content length in characters")
    remove_image: bool = Field(True, description="Remove image from content")
    headers: Dict[str, str] = None
    def execute(self, tool_input: ToolInput):
        if tool_input.get_data("input") is None:
            return None
        input = tool_input.get_data("input")
        
        # Update optional configurations
        if tool_input.get_data("mode") is not None:
            self.mode = tool_input.get_data("mode")
        if tool_input.get_data("timeout") is not None:
            self.timeout = tool_input.get_data("timeout")
        if tool_input.get_data("api_key") is not None:
            self.api_key = tool_input.get_data("api_key")
        if tool_input.get_data("max_read_content_length") is not None:
            self.max_read_content_length = tool_input.get_data("max_read_content_length")
        if tool_input.get_data("remove_image") is not None:
            self.remove_image = tool_input.get_data("remove_image")
        if tool_input.get_data("headers") is not None:
            self.headers = tool_input.get_data("headers")

        # print(f"mode: {self.mode}, max_read_content_length: {self.max_read_content_length}, headers: {self.headers}")

        if self.mode == "read":
            if input.startswith("https://") or input.startswith("http://"):
                return self.read_url(url=html.unescape(input))
            else:
                return None
        elif self.mode == "search":
            return self.search_query(query=input)
        elif self.mode == "check_fact":
            return self.check_fact(query=input)
        else:
            return None

    def _make_api_request(self, url: str, timeout: int, error_prefix: str) -> Optional[dict]:
        """
        Unified method for handling API requests
        
        Args:
            url: The URL to request
            timeout: Request timeout in seconds
            error_prefix: Error message prefix
            
        Returns:
            Optional[dict]: Response data, returns None if failed
        """
        retries = 3
        for attempt in range(retries):
            try:
                response = requests.get(
                    url, 
                    headers=self._get_headers(), 
                    verify=False, 
                    timeout=timeout
                )
                response.raise_for_status()
                content = response.json()

                if content.get("code") != 200:
                    print(f"Request failed with status code {content.get('code')}")
                    return None
                    
                return content
                
            except requests.HTTPError as e:
                error_msg = (f"Access forbidden. Please check your API key and permissions. Error: {str(e)}" 
                           if e.response.status_code == 403 
                           else f"HTTP Error: {str(e)}")
                if attempt < retries - 1:
                    time.sleep(2 ** attempt)
                    continue
                return error_msg
            except requests.Timeout:
                error_msg = f"{error_prefix}: Request timeout"
                if attempt < retries - 1:
                    time.sleep(2 ** attempt)
                    continue
                return error_msg
            except Exception as e:
                error_msg = f"{error_prefix}: {str(e)}"
                return error_msg

    def read_url(self, url: str) -> str:
        """
        Read URL and return full webpage content
        
        Args:
            url: URL to read
            
        Returns:
            str: Webpage content or error message
        """
        full_url = f"{jina_read_url}{url}"
        
        content = self._make_api_request(full_url, self.timeout, "Error reading URL")
        if isinstance(content, str):  # Error message
            return content
        if not content:
            return None     
        
        data = content.get("data", {})
        content_text = data.get("content", "").strip()
        
        if self.remove_image:
            content_text = self._remove_images(content_text)
        # Remove blank lines
        content_text = "\n".join(line for line in content_text.splitlines() if line.strip())
        return self._truncate_content(content_text)

    def search_query(self, query: str) -> str:
        """
        Execute search query and return json list string
        return example:
        [
            {
            "title": "About Jina AI",
            "description": "These photos include our former colleagues and interns—we appreciate every one of them. favorite · <strong>Founded</strong> in 2020, <strong>Jina</strong> <strong>AI</strong> is a leading search <strong>AI</strong> company.",
            "url": "https://jina.ai/about-us/",
            "content": "About Jina AI\n===============\n  \n\n[](https://jina.ai/)\n\n_search__reorder_\n\n[News](https://jina.ai/news)[Models](https://jina.ai/models)",
            "usage": {
                "tokens": 1758
                }
            },
            {
            "title": "Dr. Han Xiao, CEO and Founder of Jina AI",
            "description": "Our recent release of the jina-embeddings-v2, on <strong>October 26th</strong>, stands as a testament to Jina AI&#x27;s commitment to spearheading innovation in the AI realm. ",
            "url": "https://ai-berlin.com/blog/article/interview-with-dr-han-xiao-ceo-and-co-founder-of-jina-ai",
            "content": "Dr. Han Xiao, CEO and Founder of Jina AI | Artificial Intelligence in Berlin\n===============                \n\n[](https://ai-berlin.com/)",
            "usage": {
                "tokens": 9647
                }
            }
        ]
        """
        full_url = f"{jina_search_url}{query}"
        
        content = self._make_api_request(full_url, self.timeout, "Error executing search")
        if isinstance(content, str): 
            return content
        if not content:
            return None
            
        return str(content.get("data"))
    
    def check_fact(self, query: str) -> str:
        """
        Execute check fact query and return json string
        return example:
        {
            "factuality": 0,
            "result": false,
            "reason": "The statement is incorrect as it asserts that President Obama repealed the Smith-Mundt Act. However, the Smith-Mundt Modernization Act of 2012 did not repeal the original act. Instead, it lifted restrictions that prevented U.S. government-produced information from being disseminated to domestic audiences. The original act still exists, but its applicability has been changed. Multiple references confirm that the Smith-Mundt Act remains in effect, emphasizing that only certain restrictions were modified rather than a complete repeal.",
            "references": [
            {
                "url": "https://apnews.com/article/archive-fact-checking-7064410002",
                "keyQuote": "THE FACTS: A post circulating on Facebook with a photo of Obama falsely states he repealed a ban on government propaganda in the U.S. when he signed the National Defense Authorization Act in 2013. The amendment did not repeal the Smith-Mundt Act",
                "isSupportive": false
            },
            {
                "url": "https://en.wikipedia.org/wiki/Smith–Mundt_Act",
                "keyQuote": "The original version of the Act was amended by the Smith–Mundt Modernization Act of 2012 which allowed for materials produced by the State Department and the Broadcasting Board of Governors ...",
                "isSupportive": false
            }
            ],
            "usage": {
            "tokens": 11848
            }
        }
        """

        full_url = f"{jina_check_fact_url}{query}"
        # check_fact is very slow, so we set a longer timeout
        content = self._make_api_request(full_url, self.timeout * 3, "Error executing check fact")
        if isinstance(content, str): 
            return content
        if not content:
            return None
        return str(content.get("data"))

    def _remove_images(self, content: str) -> str:
        """
        Remove image descriptions from content
        
        Args:
            content: Text content containing image descriptions
            
        Returns:
            str: Content with image descriptions removed
        """
        return "\n".join(
            line for line in content.split("\n") 
            if not line.startswith("![Image")
        ).rstrip()

    def _get_headers(self) -> Dict[str, str]:
        """
        Get request headers with API key and other configurations
        
        Returns:
            Dict[str, str]: Headers dictionary containing Accept, X-Engine, X-Retain-Images 
            and optionally Authorization if API key is provided
        """
        headers = {
            "Accept": "application/json",
            "X-Engine": "direct",
            "X-Retain-Images": "none",
        }
        if self.headers:
            headers.update(self.headers)
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        return headers

    def _truncate_content(self, content: str) -> str:
        """
        Truncate content to the maximum allowed length.
        
        Args:
            content: Original content text
            
        Returns:
            str: Truncated content with ellipsis if exceeds max length, otherwise original content
        """
        if len(content) > self.max_read_content_length:
            truncated = content[: self.max_read_content_length]
            return truncated + "... (Content truncated)"
        return content


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: google_search_tool.py

from agentuniverse.agent.action.tool.tool import Tool, ToolInput


class MockSearchTool(Tool):
    """The demo google search mock tool.
    """

    def execute(self, tool_input: ToolInput):
        input = tool_input.get_data("input")
        # get top10 results from mock search.
        res = self.mock_api_res()
        return res

    def mock_api_res(self):
        res = f"""采访中谈及第十次减持比亚迪，巴菲特称是为了便于伯克希尔进行更好的资金配置。而在今年2月，
        芒格在美国报纸和软件公司Daily Journal举行的虚拟年会上也曾谈及减持比亚迪的 ... 
        对比亚迪的减持可能反映出巴菲特对中国新能源汽车市场未来增长的预期调整，或是对比亚迪本身估值的重新评估。
        这也可能暗示着巴菲特认为当前比亚迪股价已经 ... 
        巴菲特从2022年8月底开始首次减持比亚迪，那个时候市场就有声音表示巴菲特要有计划有节奏地清仓比亚迪了。
        差不多两年的时间，伯克希尔哈撒韦对比亚迪的持仓 ... 
        究其原因，除了前文提到的比亚迪估值偏高以及巴菲特“减持效应”之外，当前宏观经济环境的变化也是关键影响因子。
        近期美联储释放鹰派加息信号，对全球成长板块 ...
         巴菲特持续减持什么原因？比亚迪回应：公司正在和对方沟通目前经营状况良好. 
         对此，华尔街分析表示：“从调仓动作来看，巴菲特正在从科技板块撤退，回归此前熟悉的行业。
         其中最明显的变化就是巴菲特又重新大手笔建仓买入安达保险。”. 
         公司战略：巴菲特曾表示，伯克希尔哈撒韦公司希望将更多的注意力放在美国市场，这可能是其减持在美国以外市场的投资，如比亚迪的原因之一。 
         这是2023年以来，港交所第5次披露伯克希尔哈撒韦减持比亚迪H股，也是2022年8月该公司首次减持以来的第11次披露。
         至上一次披露减持后，伯克希尔哈撒韦持有比例 ... 
         业内人士认为，9月电动车行业整体渗透率提升的增速已经放缓，比亚迪三季报业绩优异，
         除了销售增速较快以外，上游碳酸锂价格下滑也是原因之一，未来业绩增速会不 ... 
         不可否认的是，巴菲特首次减持短期内可能对比亚迪股价构成压力，也有部分投资者跟风减持，但具体走势要根据企业基本面来确定。 
         事实上，在汽车业务转暖的同时，比亚迪 ...
        """
        return res



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/9 18:01
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: base_log_sink.py

from agentuniverse.base.util.logging.log_sink.base_file_log_sink import BaseFileLogSink
from agentuniverse.base.util.logging.log_type_enum import LogTypeEnum


class FlaskResponseLogSink(BaseFileLogSink):
    log_type: LogTypeEnum = LogTypeEnum.flask_response

      
    def process_record(self, record):
        record["message"] = self.generate_log(
            flask_response=record['extra'].get('flask_response'),
            elapsed_time=record['extra']['elapsed_time']
        )
        record['extra'].pop('flask_response', None)

    def generate_log(self, flask_response, elapsed_time) -> str:
        pass



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/15 11:42
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: enum.py

import enum
from enum import Enum


@enum.unique
class ConversationMessageEnum(Enum):
    INPUT = 'input'
    OUTPUT = 'output'


@enum.unique
class ConversationMessageSourceType(Enum):
    AGENT = 'agent'
    TOOL = 'tool'
    KNOWLEDGE = 'knowledge'
    LLM = 'llm'
    USER = 'user'



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/6/13 11:03
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: stream_callback.py

import asyncio
import datetime
import json
from queue import Queue
from typing import Optional, Dict, Any, Union, List
from uuid import UUID

from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.outputs import GenerationChunk, ChatGenerationChunk, LLMResult

from agentuniverse.agent.memory.conversation_memory.conversation_memory_module import ConversationMemoryModule
from agentuniverse.base.context.framework_context_manager import FrameworkContextManager


class StreamOutPutCallbackHandler(BaseCallbackHandler):
    """Callback Handler that prints to std out."""

    def __init__(self, queue_stream: asyncio.Queue, color: Optional[str] = None, agent_info: dict = None,
                 **kwargs) -> None:
        """Initialize callback handler."""
        self.queueStream = queue_stream
        self.color = color
        if agent_info is None:
            agent_info = {}
        self.agent_info = agent_info

    def on_chain_start(
            self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        return

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Print out that we finished a chain."""

    def on_agent_action(
            self, action: AgentAction, color: Optional[str] = None, **kwargs: Any
    ) -> Any:
        self.queueStream.put_nowait({
            "type": "ReAct",
            "data": {
                "output": "\nThought:" + action.log,
                "agent_info": self.agent_info
            }
        })

    def on_tool_start(
            self,
            serialized: Dict[str, Any],
            input_str: str,
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            tags: Optional[List[str]] = None,
            metadata: Optional[Dict[str, Any]] = None,
            inputs: Optional[Dict[str, Any]] = None,
            **kwargs: Any,
    ) -> Any:
        ConversationMemoryModule().add_tool_input_info(
            start_info={
                "source": self.agent_info.get('name'),
                "type": 'agent',
            },
            target=serialized.get('name'),
            params={
                "input": input_str
            },
            pair_id=f"tool_{run_id.hex}"
        )

    def on_llm_new_token(
            self,
            token: str,
            *,
            chunk: Optional[Union[GenerationChunk, ChatGenerationChunk]] = None,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any,
    ) -> Any:
        # add token chunk to the queue.
        self.queueStream.put_nowait({
            "type": "token",
            "data": {
                "chunk": chunk.text,
                "agent_info": self.agent_info
            }
        })

    def on_tool_end(
            self,
            output: str,
            color: Optional[str] = None,
            observation_prefix: Optional[str] = None,
            llm_prefix: Optional[str] = None,
            **kwargs: Any,
    ) -> None:
        """If not the final action, print out observation."""
        if observation_prefix is not None:
            self.queueStream.put_nowait({
                "type": "ReAct",
                "data": {
                    "output": '\n' + observation_prefix + output,
                    "agent_info": self.agent_info
                }
            })
        else:
            self.queueStream.put_nowait({
                "type": "ReAct",
                "data": {
                    "output": '\n Observation:' + output,
                    "agent_info": self.agent_info
                }
            })
        ConversationMemoryModule().add_tool_output_info(
            start_info={
                "source": self.agent_info.get('name'),
                "type": 'agent',
            },
            target=kwargs.get('name'),
            params={
                "output": output
            },
            pair_id=f"tool_{kwargs.get('run_id').hex}"
        )

    def on_text(
            self,
            text: str,
            color: Optional[str] = None,
            end: str = "",
            **kwargs: Any,
    ) -> None:
        """Run when agent ends."""

    def on_agent_finish(
            self, finish: AgentFinish, color: Optional[str] = None, **kwargs: Any
    ) -> None:
        """Run on agent end."""
        self.queueStream.put_nowait({
            "type": "ReAct",
            "data": {
                "output": '\nThought:' + finish.output,
                "agent_info": self.agent_info
            }
        })


class OpenAIProtocolStreamOutPutCallbackHandler(BaseCallbackHandler):
    """Callback Handler that prints to std out."""

    def __init__(self, queue_stream: asyncio.Queue, color: Optional[str] = None, agent_info: dict = None,
                 **kwargs) -> None:
        """Initialize callback handler."""
        self.queueStream = queue_stream
        self.color = color
        if agent_info is None:
            agent_info = {}
        self.agent_info = agent_info

    def add_output_stream(self, output_stream: Queue, agent_output: str) -> None:
        if not output_stream:
            return
        output = {
            "object": "chat.completion.chunk",
            "id": FrameworkContextManager().get_context('trace_id'),
            "created": int(datetime.datetime.now().timestamp()),
            "model": self.agent_info.get('name'),
            "choices": [
                {
                    "delta": {
                        "role": "assistant",
                        "content": agent_output
                    },
                    "index": 0,
                }
            ]
        }
        output_stream.put(json.dumps(output))

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Print out that we finished a chain."""

    # def on_agent_action(
    #         self, action: AgentAction, color: Optional[str] = None, **kwargs: Any
    # ) -> Any:
    #
    #     self.add_output_stream(self.queueStream, "\nThought:" + action.log)


    def on_tool_start(
            self,
            serialized: Dict[str, Any],
            input_str: str,
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            tags: Optional[List[str]] = None,
            metadata: Optional[Dict[str, Any]] = None,
            inputs: Optional[Dict[str, Any]] = None,
            **kwargs: Any,
    ) -> Any:
        ConversationMemoryModule().add_tool_input_info(
            start_info={
                "source": self.agent_info.get('name'),
                "type": 'agent',
            },
            target=serialized.get('name'),
            params={
                "input": input_str
            },
            pair_id=f"tool_{run_id.hex}"
        )


    def on_llm_new_token(
            self,
            token: str,
            *,
            chunk: Optional[Union[GenerationChunk, ChatGenerationChunk]] = None,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any,
    ) -> Any:
        self.add_output_stream(self.queueStream, token)

    def on_tool_end(
            self,
            output: str,
            color: Optional[str] = None,
            observation_prefix: Optional[str] = None,
            llm_prefix: Optional[str] = None,
            **kwargs: Any,
    ) -> None:
        """If not the final action, print out observation."""
        if observation_prefix is not None:
            self.add_output_stream(self.queueStream,'\n' + observation_prefix + output +"\n Thought: ")
        else:
            self.add_output_stream(self.queueStream, '\n Observation: ' + output +"\n Thought: ")
        ConversationMemoryModule().add_tool_output_info(
            start_info={
                "source": self.agent_info.get('name'),
                "type": 'agent',
            },
            target=kwargs.get('name'),
            params={
                "output": output
            },
            pair_id=f"tool_{kwargs.get('run_id').hex}"
        )

    def on_text(
            self,
            text: str,
            color: Optional[str] = None,
            end: str = "",
            **kwargs: Any,
    ) -> None:
        """Run when agent ends."""

    # def on_agent_finish(
    #         self, finish: AgentFinish, color: Optional[str] = None, **kwargs: Any
    # ) -> None:
    #     """Run on agent end."""
    #     self.add_output_stream(self.queueStream, '\nThought:' + finish.output)


class InvokeCallbackHandler(BaseCallbackHandler):
    """Callback Handler that prints to std out."""
    source: str
    llm_name: str

    def __init__(self, source: str, llm_name: str) -> None:
        """Initialize callback handler."""
        self.source = source
        self.llm_name = llm_name

    def on_llm_start(
            self,
            serialized: Dict[str, Any],
            prompts: List[str],
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            tags: Optional[List[str]] = None,
            metadata: Optional[Dict[str, Any]] = None,
            **kwargs: Any,
    ) -> Any:
        prompt = "\n".join(prompts)

        start_info = {
            "source": self.source,
            "type": "agent",
        }

        ConversationMemoryModule().add_llm_input_info(start_info, self.llm_name, prompt, f"llm_{run_id.hex}")

    def on_llm_end(
            self,
            response: LLMResult,
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any,
    ) -> Any:
        start_info = {
            "source": self.source,
            "type": "agent",
        }
        ConversationMemoryModule().add_llm_output_info(
            start_info, self.llm_name,
            response.generations[0][0].text,
            f"llm_{run_id.hex}"
        )



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
from typing import List

# @Time    : 2024/7/31 16:19
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: recursive_character_text_splitter.py
from typing import List, Optional
from langchain.text_splitter import RecursiveCharacterTextSplitter as Splitter

from agentuniverse.agent.action.knowledge.doc_processor.doc_processor import \
    DocProcessor
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class RecursiveCharacterTextSplitter(DocProcessor):
    """Splits text recursively using a hierarchy of character separators."""
    chunk_size: int = 200
    chunk_overlap: int = 20
    separators: List[str] = ["\n\n", "\n", " ", ""]
    __splitter: Optional[Splitter] = None

    @property
    def splitter(self) -> Splitter:
        if not self.__splitter:
            self.__splitter = Splitter(separator=self.separator,
                                       chunk_size=self.chunk_size,
                                       chunk_overlap=self.chunk_overlap)
        return self.__splitter

    def _process_docs(self, origin_docs: List[Document], query: Query = None) -> \
            List[Document]:
        """Split documents recursively using character separators.
        
        Args:
            origin_docs: List of documents to be split.
            query: Optional query object (not used in this processor).
            
        Returns:
            List of split document chunks.
        """
        lc_doc_list = self.splitter.split_documents(Document.as_langchain_list(
            origin_docs
        ))
        return Document.from_langchain_list(lc_doc_list)

    def _initialize_by_component_configer(self,
                                         doc_processor_configer: ComponentConfiger) -> 'DocProcessor':
        """Initialize splitter parameters from configuration.
        
        Args:
            doc_processor_configer: Configuration object containing splitter parameters.
            
        Returns:
            Initialized document processor instance.
        """
        super()._initialize_by_component_configer(doc_processor_configer)
        if hasattr(doc_processor_configer, "chunk_size"):
            self.chunk_size = doc_processor_configer.chunk_size
        if hasattr(doc_processor_configer, "chunk_overlap"):
            self.chunk_overlap = doc_processor_configer.chunk_overlap
        if hasattr(doc_processor_configer, "separators"):
            self.separators = doc_processor_configer.separators
        return self










# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/2 16:18
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: llm_channel_manager.py
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase


@singleton
class LLMChannelManager(ComponentManagerBase):

    def __init__(self):
        super().__init__(ComponentEnum.LLM_CHANNEL)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/2/2 22:00
# @Author  : wangyapei
# @FileName: csv_reader.py

import csv
from pathlib import Path
from typing import List, Union, Optional, Dict

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document


class CSVReader(Reader):
    """CSV file reader.
    
    Used to read and parse CSV format files, supports local file paths or file objects as input.
    """

    def _load_data(self, 
                  file: Union[str, Path], 
                  delimiter: str = ",",
                  quotechar: str = '"',
                  ext_info: Optional[Dict] = None) -> List[Document]:
        """Parse CSV file.

        Args:
            file: CSV file path or file object
            delimiter: CSV delimiter, default is comma
            quotechar: Quote character, default is double quote
            ext_info: Additional metadata information

        Returns:
            List[Document]: List of documents containing CSV content

        Raises:
            FileNotFoundError: Raised when file does not exist
            ValueError: Raised when file reading fails
        """
        try:
            if isinstance(file, str):
                file = Path(file)
            
            if isinstance(file, Path):
                if not file.exists():
                    raise FileNotFoundError(f"File not found: {file}")
                file_content = file.open(newline="", mode="r", encoding="utf-8")
            else:
                file.seek(0)
                file_content = file

            csv_content = []
            with file_content as csvfile:
                csv_reader = csv.reader(csvfile, delimiter=delimiter, quotechar=quotechar)
                for row in csv_reader:
                    # Filter out completely empty rows
                    if any(cell.strip() for cell in row):
                        # Remove empty values at the end of row
                        while row and not row[-1].strip():
                            row.pop()
                        # Only add non-empty values to result
                        csv_content.append(", ".join(filter(None, row)))
            
            # Combine all valid rows into final text
            final_content = "\n".join(csv_content)

            # Get metadata
            metadata = {"file_name": getattr(file, 'name', 'unknown')}
            if ext_info:
                metadata.update(ext_info)
            # print(f"csv_content: {final_content} \n metadata: {metadata}")
            return [Document(text=final_content, metadata=metadata)]
        except Exception as e:
            raise ValueError(f"Failed to read CSV file: {str(e)}") from e



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/9/29 15:51
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agent_template.py
from abc import ABC
from typing import Optional
from queue import Queue

from langchain_core.output_parsers import StrOutputParser

from agentuniverse.agent.agent import Agent
from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.base.util.agent_util import assemble_memory_input, assemble_memory_output
from agentuniverse.base.util.memory_util import get_memory_string
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.llm.llm import LLM
from agentuniverse.prompt.chat_prompt import ChatPrompt
from agentuniverse.prompt.prompt import Prompt


class AgentTemplate(Agent, ABC):
    llm_name: Optional[str] = ''
    memory_name: Optional[str] = None
    tool_names: Optional[list[str]] = None
    knowledge_names: Optional[list[str]] = None
    prompt_version: Optional[str] = None
    conversation_memory_name: Optional[str] = None

    def execute(self, input_object: InputObject, agent_input: dict, **kwargs) -> dict:
        memory: Memory = self.process_memory(agent_input, **kwargs)
        llm: LLM = self.process_llm(**kwargs)
        prompt: Prompt = self.process_prompt(agent_input, **kwargs)
        return self.customized_execute(input_object, agent_input, memory, llm, prompt, **kwargs)

    async def async_execute(self, input_object: InputObject, agent_input: dict, **kwargs) -> dict:
        memory: Memory = self.process_memory(agent_input, **kwargs)
        llm: LLM = self.process_llm(**kwargs)
        prompt: Prompt = self.process_prompt(agent_input, **kwargs)
        return await self.customized_async_execute(input_object, agent_input, memory, llm, prompt, **kwargs)

    def customized_execute(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM, prompt: Prompt,
                           **kwargs) -> dict:
        self.load_memory(memory, agent_input)
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile, agent_input)
        chain = prompt.as_langchain() | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        res = self.invoke_chain(chain, agent_input, input_object, **kwargs)
        self.add_memory(memory, f"Human: {agent_input.get('input')}, AI: {res}", agent_input=agent_input)
        self.add_output_stream(input_object.get_data('output_stream'), res)
        return {**agent_input, 'output': res}

    async def customized_async_execute(self, input_object: InputObject, agent_input: dict, memory: Memory,
                                       llm: LLM, prompt: Prompt, **kwargs) -> dict:
        assemble_memory_input(memory, agent_input, self.get_memory_params(agent_input))
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile, agent_input)
        chain = prompt.as_langchain() | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        res = await self.async_invoke_chain(chain, agent_input, input_object, **kwargs)
        if self.memory_name:
            assemble_memory_output(memory=memory,
                                   agent_input=agent_input,
                                   content=f"Human: {agent_input.get('input')}, AI: {res}")
        self.add_output_stream(input_object.get_data('output_stream'), res)
        return {**agent_input, 'output': res}

    def validate_required_params(self):
        pass

    def add_output_stream(self, output_stream: Queue, agent_output: str) -> None:
        pass

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'AgentTemplate':
        super().initialize_by_component_configer(component_configer)
        self.llm_name = self.agent_model.profile.get('llm_model', {}).get('name')
        self.memory_name = self.agent_model.memory.get('name')
        self.tool_names = self.agent_model.action.get('tool', [])
        self.knowledge_names = self.agent_model.action.get('knowledge', [])
        self.conversation_memory_name = self.agent_model.memory.get('conversation_memory', '')
        return self

    def process_llm(self, **kwargs) -> LLM:
        return super().process_llm(llm_name=self.llm_name)

    def process_memory(self, agent_input: dict, **kwargs) -> Memory | None:
        if 'chat_history' in agent_input and agent_input.get('chat_history'):
            if isinstance(agent_input.get('chat_history'), list):
                agent_input['chat_history'] = get_memory_string(agent_input.get('chat_history'),
                                                                agent_input.get('agent_id'))
                return None
            else:
                return None
        return super().process_memory(agent_input=agent_input, memory_name=self.memory_name, llm_name=self.llm_name)

    def invoke_tools(self, input_object: InputObject, **kwargs) -> str:
        return super().invoke_tools(input_object=input_object, tool_names=self.tool_names)

    def invoke_knowledge(self, query_str: str, input_object: InputObject, **kwargs) -> str:
        return super().invoke_knowledge(query_str=query_str, input_object=input_object,
                                        knowledge_names=self.knowledge_names)

    def process_prompt(self, agent_input: dict, **kwargs) -> ChatPrompt:
        return super().process_prompt(agent_input=agent_input, prompt_version=self.prompt_version)

    def create_copy(self) -> 'AgentTemplate':
        copied = super().create_copy()
        copied.llm_name = self.llm_name
        copied.memory_name = self.memory_name
        copied.tool_names = self.tool_names.copy() if self.tool_names is not None else None
        copied.knowledge_names = self.knowledge_names.copy() if self.knowledge_names is not None else None
        copied.prompt_version = self.prompt_version
        copied.conversation_memory_name = self.conversation_memory_name
        return copied



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/17 11:51
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: slave_rag_agent_template.py
# !/usr/bin/env python3
# -*- coding:utf-8 -*-
from typing import Any
from langchain_core.runnables import RunnableSerializable
import datetime
import re

from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.agent.input_object import InputObject
from agentuniverse.prompt.chat_prompt import ChatPrompt
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.prompt.prompt_manager import PromptManager
from agentuniverse.prompt.prompt_model import AgentPromptModel
from agentuniverse.base.context.context_archive_utils import get_current_context_archive


class SlaveRagAgentTemplate(AgentTemplate):

    def input_keys(self) -> list[str]:
        return ['prompt_name', 'prompt_params']

    def output_keys(self) -> list[str]:
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:

        agent_input['prompt_name'] = input_object.get_data('prompt_name')
        agent_input['prompt_params'] = input_object.get_data('prompt_params')
        context_archive = get_current_context_archive()

        # get archive data from context
        agent_input.update(agent_input.get('prompt_params'))
        for k, v in agent_input.items():
            result = v
            if isinstance(v, str):
                pattern = r'\$\#\{(\w+)\}'
                result = re.sub(pattern, lambda match: context_archive.get(match.group(1), {}).get('data', match.group(0)), v)
            elif isinstance(v, list):
                replaced_str = []
                for origin_str in v:
                    pattern = r'\$\#\{(\w+)\}'
                    replaced_str.append(re.sub(pattern, lambda match: context_archive.get(
                        match.group(1), {}).get('data', match.group(0)), origin_str))
                result = '\n****************************\n'.join(replaced_str)

            agent_input[k] = result
            input_object.add_data(k, result)
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        return {**agent_result, 'output': agent_result['output']}

    def process_prompt(self, agent_input: dict, **kwargs) -> ChatPrompt:
        expert_framework = agent_input.pop('expert_framework', '') or ''

        profile: dict = self.agent_model.profile

        profile_instruction = profile.get('instruction')
        profile_instruction = expert_framework + profile_instruction if profile_instruction else profile_instruction

        profile_prompt_model: AgentPromptModel = AgentPromptModel(introduction=profile.get('introduction'),
                                                                  target=profile.get('target'),
                                                                  instruction=profile_instruction)
        # get the prompt by the prompt version
        version_prompt: Prompt = PromptManager().get_instance_obj(agent_input['prompt_name'])

        if version_prompt is None and not profile_prompt_model:
            raise Exception("Either the `prompt_version` or `introduction & target & instruction`"
                            " in agent profile configuration should be provided.")
        if version_prompt:
            version_prompt_model: AgentPromptModel = AgentPromptModel(
                introduction=getattr(version_prompt, 'introduction', ''),
                target=getattr(version_prompt, 'target', ''),
                instruction=expert_framework + getattr(version_prompt, 'instruction', ''))
            profile_prompt_model = profile_prompt_model + version_prompt_model

        chat_prompt = ChatPrompt().build_prompt(profile_prompt_model, ['introduction', 'target', 'instruction'])
        image_urls: list = agent_input.pop('image_urls', []) or []
        if image_urls:
            chat_prompt.generate_image_prompt(image_urls)
        return chat_prompt



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/02/14 23:05
# @Author  : zhouxiaoji
# @Email   : zh_xiaoji@qq.com
# @FileName: image_reader.py

from pathlib import Path
from typing import Any, List, Optional, Dict, Union
import numpy as np
from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger

SUPPORTED_FORMATS = {'jpg', 'jpeg', 'png', 'bmp', 'webp', 'tiff'}


class ImageReader(Reader):
    """Image reader."""

    languages: List[str] = ['en', 'ch_sim']
    use_gpu: bool = False
    image_model: str = "microsoft/git-base-coco"
    _ocr_reader = None
    _description_pipeline = None

    def _initialize_by_component_configer(
            self, reader_configer: ComponentConfiger) -> 'ImageReader':
        """Initialize the reader by the ComponentConfiger object.

        Args:
            reader_configer: A configer contains reader basic info.
        Returns:
            ImageReader: A reader instance.
        """
        super()._initialize_by_component_configer(reader_configer)
        if hasattr(reader_configer, "languages"):
            self.languages = reader_configer.languages
        if hasattr(reader_configer, "use_gpu"):
            self.use_gpu = reader_configer.use_gpu
        if hasattr(reader_configer, "image_model"):
            self.image_model = reader_configer.image_model
        return self

    def _get_ocr_reader(self):
        """Lazy initialization of OCR reader."""
        if self._ocr_reader is None:
            try:
                import easyocr
            except ImportError:
                raise ImportError(
                    "easyocr is required. Install with: pip install easyocr")
            self._ocr_reader = easyocr.Reader(self.languages, gpu=self.use_gpu)
        return self._ocr_reader

    def _get_description_pipeline(self):
        """Lazy initialization of image description pipeline."""
        if self._description_pipeline is None:
            try:
                from transformers import pipeline
            except ImportError:
                raise ImportError(
                    "transformers is required. Install with: pip install transformers"
                )
            self._description_pipeline = pipeline("image-to-text",
                                                  model=self.image_model)
        return self._description_pipeline

    def extract_text(self, image: Any) -> str:
        """Extract text from image using OCR."""
        reader = self._get_ocr_reader()
        results = reader.readtext(np.array(image))
        return "\n".join([text[1] for text in results])

    def generate_description(self, image: Any) -> str:
        """Generate description of the image content."""
        pipeline = self._get_description_pipeline()
        description = pipeline(image)[0]['generated_text']
        return description

    def _load_data(self,
                   file: Union[str, Path],
                   ext_info: Optional[Dict] = None) -> List[Document]:
        """Load and process image file to extract text content.

        Args:
            file (Union[str, Path]): Path to the image file to be processed
            ext_info (Optional[Dict], optional): Additional metadata or processing parameters. Defaults to None.

        Returns:
            List[Document]: A list of Document objects containing the extracted text and metadata.

        Note:
            Current implementation converts image to text before embedding to maintain vector space
            consistency and semantic alignment with other document types in the knowledge base.
            However, a more optimal approach would be direct image-to-embedding conversion, which
            could better preserve visual semantic information and potentially improve retrieval
            quality for image-related queries.
        """
        try:
            from PIL import Image, ImageFile
        except ImportError as e:
            raise ImportError(
                "PIL is required. Install with: pip install Pillow") from e

        def parse_metadata(image: ImageFile) -> Dict[str, Any]:
            width, height = image.size
            channels = len(image.getbands())
            mode = image.mode
            metadata: Dict[str, Any] = {
                'file_name': file.name,
                'format': image_format,
                'image_size': {
                    'width': width,
                    'height': height
                },
                'channel': channels,
                'color_mode': mode
            }
            if ext_info is not None:
                metadata.update(ext_info)
            return metadata

        try:
            if isinstance(file, str):
                file = Path(file)
            if not file.exists():
                raise FileNotFoundError(f"Image file not found: {file}")

            image_format = file.suffix.lower().lstrip('.')
            if image_format not in SUPPORTED_FORMATS:
                raise ValueError(
                    f"Unsupported image format: {image_format}. "
                    f"Supported formats are: {', '.join(SUPPORTED_FORMATS)}")

            image: ImageFile = Image.open(file)
            ocr_text: str = self.extract_text(image)
            image_description: str = self.generate_description(image)
            image_text = f"{ocr_text}\n{image_description}" if (
                ocr_text.strip() and image_description.strip()
            ) else f"{ocr_text}{image_description}"

            metadata = parse_metadata(image)
            metadata.update({
                'has_text': bool(ocr_text.strip()),
                'has_description': bool(image_description.strip())
            })
            return [Document(text=image_text, metadata=metadata)]
        except Exception as e:
            raise Exception(f"Error processing image {file}: {str(e)}")



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: python_repl.py

import re

from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from langchain_community.utilities import PythonREPL
from pydantic import Field


class PythonREPLTool(Tool):
    """The mock search tool.

    In this tool, we mocked the search engine's answers to search for information about BYD and Warren Buffett.

    Note:
        The tool is only suitable for users searching for Buffett or BYD related queries.
        We recommend that you configure your `SERPER_API_KEY` and use google_search_tool to get information.
    """
    client: PythonREPL = Field(default_factory=lambda: PythonREPL())

    def execute(self, tool_input: ToolInput):
        input = tool_input.get_data("input")
        """Demonstrates the execute method of the Tool class."""
        pattern = re.compile(r"```python(.*?)``", re.DOTALL)
        matches = pattern.findall(input)
        if len(matches) == 0:
            pattern = re.compile(r"```py(.*?)``", re.DOTALL)
            matches = pattern.findall(input)
        if len(matches) == 0:
            return self.client.run(input)
        res = self.client.run(matches[0])
        if res == "" or res is None:
            return "ERROR: 你的python代码中没有使用print输出任何内容，请参考工具示例"
        else:
            return res



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/02/15 23:35
# @Author  : zhouxiaoji
# @Email   : zh_xiaoji@qq.com
# @FileName: doubao_embedding.py

from typing import Any, List, Optional
from pydantic import Field
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.agent.action.knowledge.embedding.embedding import Embedding
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger

SUPPORTED_DIMENSIONS = {512, 1024, 2048}


class DoubaoEmbedding(Embedding):
    """The Doubao embedding class using Volcengine Ark Runtime."""

    ark_api_key: Optional[str] = Field(
        default_factory=lambda: get_from_env("ARK_API_KEY"))
    endpoint_id: Optional[str] = Field(
        default_factory=lambda: get_from_env("ARK_ENDPOINT_ID"))
    client: Optional[Any] = None
    embedding_dims: Optional[int] = None

    def get_embeddings(self, texts: List[str], **kwargs) -> List[List[float]]:
        """
        Retrieve text embeddings for a list of input texts using Doubao API.

        Args:
            texts (List[str]): A list of input texts to be embedded.

        Returns:
            List[List[float]]: A list of embeddings corresponding to the input texts.

        Raises:
            Exception: If the API call fails or if required configuration is missing.
        """
        if not self.client:
            if not self.ark_api_key:
                raise ValueError("ARK_API_KEY is required but not set")
            try:
                from volcenginesdkarkruntime import Ark
            except ImportError:
                raise ImportError(
                    "Ark is required. Install with: pip install volcengine-python-sdk[ark]"
                )
            self.client = Ark(api_key=self.ark_api_key)
        if not self.endpoint_id:
            raise ValueError("endpoint_id is required but not set")

        def sliced_norm_l2(vec: List[float]) -> List[float]:
            if self.embedding_dims not in SUPPORTED_DIMENSIONS:
                raise ValueError(
                    f"Unsupported embedding dimension: {self.embedding_dims}. "
                    f"Supported dimensions are: {', '.join(SUPPORTED_DIMENSIONS)}"
                )
            import numpy as np
            norm = float(np.linalg.norm(vec[:self.embedding_dims]))
            return [v / norm for v in vec[:self.embedding_dims]]

        try:
            response = self.client.embeddings.create(model=self.endpoint_id,
                                                     input=texts)
            if self.embedding_dims is None:
                return [data.embedding for data in response.data]
            return [sliced_norm_l2(data.embedding) for data in response.data]
        except Exception as e:
            raise Exception(
                f"Failed to get embedding from Doubao API: {str(e)}")

    async def async_get_embeddings(self, texts: List[str],
                                   **kwargs) -> List[List[float]]:
        return self.get_embeddings(texts)

    def _initialize_by_component_configer(
            self, embedding_configer: ComponentConfiger) -> 'Embedding':
        """Initialize the embedding by the ComponentConfiger object.

        Args:
            embedding_configer(ComponentConfiger): A configer contains embedding configuration.
        Returns:
            Embedding: A DoubaoEmbedding instance.
        """
        super()._initialize_by_component_configer(embedding_configer)
        if hasattr(embedding_configer, "ark_api_key"):
            self.ark_api_key = embedding_configer.ark_api_key
        if hasattr(embedding_configer, "endpoint_id"):
            self.endpoint_id = embedding_configer.endpoint_id
        if hasattr(embedding_configer, "embedding_dims"):
            self.embedding_dims = embedding_configer.embedding_dims
        if self.ark_api_key:
            try:
                from volcenginesdkarkruntime import Ark
            except ImportError:
                raise ImportError(
                    "Ark is required. Install with: pip install volcengine-python-sdk[ark]"
                )
            self.client = Ark(api_key=self.ark_api_key)
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/03/04 13:19
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: code_reader.py
import os
from pathlib import Path
from typing import List, Union, Optional, Dict

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document


CODE_FILE_EXTENSIONS = {
    '.py': 'python',
    '.js': 'javascript',
    '.ts': 'typescript',
    '.java': 'java',
    '.cpp': 'cpp',
    '.c': 'c',
    '.h': 'c',
    '.hpp': 'cpp',
    '.cs': 'csharp',
    '.go': 'go',
    '.rb': 'ruby',
    '.php': 'php',
    '.swift': 'swift',
    '.kt': 'kotlin',
    '.rs': 'rust',
    '.sh': 'shell',
    '.html': 'html',
    '.css': 'css',
    '.sql': 'sql',
    '.json': 'json',
    '.xml': 'xml',
    '.yaml': 'yaml',
    '.yml': 'yaml',
    '.md': 'markdown',
}


class CodeReader(Reader):

    def _load_data(self,
                   file: Union[str, Path],
                   ext_info: Optional[Dict] = None) -> List[Document]:
        if isinstance(file, str):
            file = Path(file)
        if isinstance(file, Path):
            if not file.exists():
                raise FileNotFoundError(f"Code file not found: {file}")
        file_content = file.read_text(encoding="utf-8")
        file_name = file.name
        file_suffix = file.suffix.lower()
        language = CODE_FILE_EXTENSIONS.get(file_suffix, 'unknown')
        metadata = {
            "file_name": file_name,
            "file_path": str(file),
            "language": language,
            "file_suffix": file_suffix,
        }
        if ext_info:
            metadata.update(ext_info)
        return [Document(text=file_content, metadata=metadata)]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/4 15:06
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: code_types.py

from typing import Dict, Optional, TypedDict

from agentuniverse.agent.action.knowledge.doc_processor.types.ast_types import AstNode
from agentuniverse.agent.action.knowledge.doc_processor.types.metrics_types import CodeMetrics


class CodeFeatures(TypedDict):
    node_counts: Dict[str, int]
    code_metrics: CodeMetrics
    identifier_count: int
    function_count: int
    class_count: int
    statement_count: int


class CodeRepresentation(TypedDict):
    ast: AstNode
    features: CodeFeatures
    language: str
    code_length: int


class ChunkRepresentation(TypedDict):
    ast: AstNode
    code: str
    language: str
    name: Optional[str]
    type: str



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/2/28 17:46
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: retry.py

import functools
import time
from typing import Any, Callable


def retry(max_retries: int = 3, delay: float = 1.0) -> Callable:
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            retries = 0
            last_exception = None
            while retries < max_retries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    last_exception = e
                    retries += 1
                    if retries < max_retries:
                        time.sleep(delay)
            raise Exception(f"Failed after {max_retries} retries. Last error: {str(last_exception)}")
        return wrapper
    return decorator



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
from typing import Union

# @Time    : 2025/1/17 12:03
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: llm_invocation_log_sink.py

from agentuniverse.base.util.logging.log_sink.base_file_log_sink import BaseFileLogSink
from agentuniverse.base.util.logging.log_type_enum import LogTypeEnum
from agentuniverse.base.util.monitor.monitor import Monitor


class LLMInvocationLogSink(BaseFileLogSink):
    log_type: LogTypeEnum = LogTypeEnum.llm_invocation

    def process_record(self, record):
        record["message"] = self.generate_log(
            used_token=record['extra'].get('used_token'),
            cost_time=record['extra'].get('cost_time'),
            llm_output=record['extra'].get('llm_output'),
        )

    def generate_log(self, used_token: int, cost_time: float, llm_output:  Union[str, dict]) -> str:
        log_str = f" LLM cost {cost_time:.2f} seconds"
        if used_token:
            log_str += f", token usage: {used_token}"
        return Monitor.get_invocation_chain_str() + log_str + f" LLM output finished."
